# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The module that implements the default backend of the state.
"""

r"""
Default simulations for States in QuAIRKit obeys the following broadcast rule:

| Batch size of Operators | Batch size of Input State | Batch size of Output State |
|:-----------------------:|:-------------------------:|:--------------------------:|
|           1             |             1             |             1            |
|           n             |             1             |             n            |
|           1             |             n             |             n            |
|           n             |             n             |             n            |
|           m             |             n             |          (m, n)          |

Note that the last rule currently is only applicable to the `evolve_keep_dim` method.
"""

import math
import warnings
from abc import abstractmethod
from typing import List, Optional, Tuple, Type, Union

import numpy as np
import torch

from ... import utils
from .base import State
from .simulator import StateSimulator

__all__ = ['DefaultSimulator', 'MixedState', 'PureState']


class DefaultSimulator(StateSimulator):
    r"""The abstract base class for default simulators of quantum states in QuAIRKit.
    """
    backend: str = 'default'
    
    @staticmethod
    def fetch_initializer(shape: List[int]) -> Union[Type['PureState'], Type['MixedState']]:
        r"""Determine whether the input data should be a (batch of) pure/mixed state in ``'default'`` backend.
        
        Note:
            When the ``backend`` is set as ``'default'``, the default initializer of this state is determined by the table
            
            +----------------+---------------------+---------------------+
            |                | single              | batch               |
            +================+=====================+=====================+
            | state_vector   | [d], [1, d], [d, 1] | [d1, ..., dn, d, 1] |
            +----------------+---------------------+---------------------+
            | density_matrix | [d, d]              | [d1, ..., dn, d, d] |
            +----------------+---------------------+---------------------+
        """
        if len(shape) == 1:
            return PureState

        if len(shape) == 2:
            return MixedState if shape[0] == shape[1] else PureState
        
        # batch case
        if shape[-1] == 1:   # Pure state batch
            return PureState
        if shape[-1] == shape[-2]:  # Mixed state batch
            return MixedState
        
        raise ValueError(
            f"The input data shape does not match the 'default' backend: received shape {shape}, "
            "expect one of [d], [1, d], [d, 1], [d, d], [..., d, 1] or [..., d, d].")
    
    def __new__(cls, data: torch.Tensor, sys_dim: List[int], 
                system_seq: Optional[List[int]], probability: Optional[List[torch.Tensor]]):
        subclass = cls.fetch_initializer(data.shape)
        instance = super(DefaultSimulator, subclass).__new__(subclass)
        if instance is not None and hasattr(instance, '__init__'):
            instance.__init__(data, sys_dim, system_seq, probability)
        return instance
        
    @abstractmethod
    def _expand(self, batch_dim: List[int]) -> 'DefaultSimulator':
        r"""Expand the batch dimension of the State.
        
        Args:
            batch_dim: the new batch dimension
            
        Note:
            For internal use only, see torch.expand() for more information about expand.
        
        """

    def expand_as(self, other: 'DefaultSimulator') -> 'DefaultSimulator':
        assert self._prob_dim == [], \
            "Does not support expanding a State with probability"
        
        other_batch_dim = other.batch_dim
        expand_state = self._expand(other_batch_dim)
        expand_state._batch_dim = other._batch_dim.copy()
        expand_state._prob = [prob.clone() for prob in other._prob]
        return expand_state

    def __str__(self) -> str:
        split_line = "\n-----------------------------------------------------\n"
        s = f"{split_line} Backend: {self.backend}\n"
        s += f" System dimension: {self._sys_dim}\n"
        s += f" System sequence: {self._system_seq}\n"

        data = np.round(self.numpy(), decimals=2)
        interrupt_num = 5
        if not self.batch_dim:
            s += str(data.squeeze(0))
            s += split_line
            return s

        s += f" Batch size: {self.batch_dim}\n"
        for i, mat in enumerate(data):
            s += f"\n # {i}:\n{mat}"
            
            if i > interrupt_num:
                break_line = ("\n----------skipped for the rest of " + 
                              f"{list(data.shape)[0] - interrupt_num} states----------\n")
                s+= break_line
                return s
        
        s += split_line
        return s

    @property
    def data(self) -> torch.Tensor:
        r"""The data tensor of this state
        """
        warnings.warn(
            'The data property is depreciated, use ket or density_matrix instead', DeprecationWarning)
        return self._data
    
    def kron(self, other: 'DefaultSimulator') -> 'DefaultSimulator':
        system_dim = self.system_dim + other.system_dim
        system_seq = self.system_seq + [x + self.num_systems for x in other.system_seq]
        
        if self._prob:
            prob = self._prob
            if other._prob:
                warnings.warn(
                    "Detect tensor product of two probabilistic states: will discard prob info of the 2nd one", UserWarning)
        else:
            prob = other._prob
            
        if self.backend == 'default-pure' and other.backend == 'default-pure':
            data = utils.linalg._kron(self.ket, other.ket)
            return PureState(data, system_dim, system_seq, prob)
        else:
            data = utils.linalg._kron(self.density_matrix, other.density_matrix)
            return MixedState(data, system_dim, system_seq, prob)
    
    @staticmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        initializer = DefaultSimulator.fetch_initializer(data.shape)
        return initializer.check(data, sys_dim, eps)
    
    def _joint_probability(self, prob_idx: List[int]) -> torch.Tensor:
        result_prob = torch.tensor(1.0)
        for idx in sorted(prob_idx):
            selected_prob = self._prob[idx]
            result_prob = result_prob.view(list(result_prob.shape) + [1] * (selected_prob.dim() - result_prob.dim()))
            result_prob = torch.mul(result_prob, selected_prob)
        return result_prob
    
    @property
    def probability(self) -> torch.Tensor:
        return self._joint_probability(range(len(self._prob)))
    
    @property
    def _prob_dim(self) -> List[int]:
        return list(self._prob[-1].shape[-len(self._prob):]) if self._prob else []
    
    @property
    def batch_dim(self) -> List[int]:
        return self._batch_dim + self._prob_dim
    
    @property
    def _squeeze_shape(self) -> List[int]:
        r"""The squeezed shape of this state batch
        """
        return [-1, int(np.prod(self._prob_dim))]
    
    @property
    def shape(self) -> torch.Size:
        return self._data.shape
    
    @State.system_dim.setter
    def system_dim(self, sys_dim: List[int]) -> None:
        self.reset_sequence()
        super(DefaultSimulator, DefaultSimulator).system_dim.__set__(self, sys_dim)
        self._system_seq = list(range(len(sys_dim)))

    @property
    def rank(self) -> List[int]:
        r"""The rank of the state.
        """
        dtype = self._data.dtype
        tol = 1e-8 if dtype == torch.complex64 else 1e-12
        tol *= self.dim
        return torch.linalg.matrix_rank(self.density_matrix, 
                                        tol=tol, hermitian=True).tolist()
    
    def numpy(self) -> np.ndarray:
        return self._data.detach().numpy()
    
    def to(self, dtype: str = None, device: str = None) -> 'DefaultSimulator':
        new_state = self.clone()
        new_state._data = new_state._data.to(dtype=dtype, device=device)
        return new_state
    
    @property
    def system_seq(self) -> List[int]:
        r"""The system order of this state
        """
        return self._system_seq.copy()

    @system_seq.setter
    @abstractmethod
    def system_seq(self, target_seq: List[int]) -> None:
        r"""Set the system order of the state.

        Args:
            target_seq: the target systems order.

        """
    
    def permute(self, target_seq: List[int]) -> 'DefaultSimulator':
        new_state = self.clone()
        new_state.system_seq = target_seq
        new_state._system_seq = list(range(self.num_systems))
        return new_state

    def reset_sequence(self) -> None:
        r"""reset the system order to default sequence i.e. from 0 to n - 1.
        """
        self.system_seq = list(range(self.num_systems))
    

class MixedState(DefaultSimulator):
    r"""The mixed state class. 

    Args:
        data: tensor array (in density matrix representation) for quantum mixed state(s).
        sys_dim: a list of dimensions for each system.
        system_seq: the system order of this state. Defaults to be from 0 to n - 1.
        probability: list of state probability distributions. Defaults to be 1.
    
    Note:
        The data is stored in the matrix-form with shape :math:`(-1, d, d)`

    """
    backend: str = 'default-mixed'
    
    def __init__(self, data: torch.Tensor, sys_dim: List[int], 
                 system_seq: Optional[List[int]] = None,
                 probability: Optional[List[torch.Tensor]] = None):
        super().__init__(data, sys_dim, system_seq, probability)

        dim = self.dim
        self._data = data.reshape([-1, dim, dim])
    
    def __getitem__(self, key: Union[int, slice]) -> 'MixedState':
        assert self.batch_dim, \
            f"This state is not batched and hence cannot be indexed: received key {key}."
        return MixedState(self.density_matrix[key], self.system_dim, self.system_seq,
                          [prob.clone() for prob in self._prob])
    
    def prob_select(self, outcome_idx: torch.Tensor, prob_idx: int = -1) -> 'MixedState':
        num_prob = len(self._prob)
        if prob_idx > 0:
            prob_idx -= len(self._prob) 

        new_prob = []
        for idx, prob in enumerate(self._prob):
            if num_prob + prob_idx > idx:
                new_prob.append(prob.clone())
            else:
                new_prob.append(prob.index_select(dim=prob_idx, index=outcome_idx).squeeze(prob_idx))

        data_idx = prob_idx - 2
        data = self.density_matrix.index_select(dim=data_idx, index=outcome_idx).squeeze(data_idx)
        return MixedState(data, self.system_dim, self.system_seq, new_prob)
    
    def _expand(self, batch_dim: List[int]) -> 'MixedState':
        if original_batch_dim := self.batch_dim:
            assert batch_dim[-len(original_batch_dim):] == original_batch_dim, \
                f"Expand dimension mismatch: expected {original_batch_dim}, received {batch_dim[-len(original_batch_dim):]}."
        
        expand_state = self.clone()
        expand_state._data = self.density_matrix.expand(batch_dim + [-1, -1]).reshape([-1, self.dim, self.dim])
        return expand_state
    
    def add_probability(self, prob: torch.Tensor) -> None:
        num_previous_prob = len(self._prob)
        prob = prob.view([1] * num_previous_prob + [-1]).to(self.dtype).real

        data = self._data.view(self.batch_dim + [1, self.dim, self.dim])
        self._data = data.repeat(self.batch_dim + [prob.shape[-1], 1, 1]).view([-1, self.dim, self.dim])
        self._prob.append(prob)
    
    @staticmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        if isinstance(sys_dim, List):
            num_sys = len(sys_dim)
            expected_dimension = int(np.prod(sys_dim))
        else:
            num_sys = int(math.log(data.shape[-2], sys_dim))
            expected_dimension = sys_dim ** num_sys
        
        assert data.shape[-2] == expected_dimension, \
            f"Input data shape {data.shape} does not match the input system dimension {sys_dim}."
        
        if eps:
            assert torch.all(is_density := utils.check._is_density_matrix(data, eps)), \
                f"(Some) data is not a density matrix: asserted {is_density.tolist()}"
        
        return num_sys

    @property
    def ket(self) -> torch.Tensor:
        raise NotImplementedError(
            "Mixed state does not support state vector representation." +
            "If you are looking for the vectorization of the mixed state, please call the 'vec' property.")

    @property
    def density_matrix(self) -> torch.Tensor:
        self.reset_sequence()
        return self._data.view(self.batch_dim + [self.dim, self.dim]).clone()
    
    def _trace(self, trace_idx: List[int]) -> 'MixedState':
        remain_seq, remain_system_dim = [], []
        for x in self._system_seq:
            if x not in trace_idx:
                remain_seq.append(x)
                remain_system_dim.append(self.system_dim[x])
        trace_dim = math.prod([self.system_dim[x] for x in trace_idx])
        self.system_seq = trace_idx + remain_seq
        
        data = self._data.view(self.batch_dim + [self.dim, self.dim])
        data = utils.linalg._trace_1(data, trace_dim)
        
        # convert remaining sequence
        value_to_index = {value: index for index, value in enumerate(sorted(remain_seq))}
        remain_seq = [value_to_index[i] for i in remain_seq]
        return MixedState(data, remain_system_dim, remain_seq, self._prob)
    
    def _reset(self, reset_idx: List[int], replace_state: 'DefaultSimulator') -> 'MixedState':
        remain_seq = [x for x in self._system_seq if x not in reset_idx]
        remain_seq.sort()
        
        trace_dim = math.prod([self.system_dim[x] for x in reset_idx])
        self.system_seq = reset_idx + remain_seq

        data = self._data.view(self.batch_dim + [self.dim, self.dim])
        data = utils.linalg._kron(utils.linalg._trace_1(data, trace_dim), replace_state.density_matrix)
        return MixedState(data, self.system_dim, self.system_seq, self._prob)
    
    def _transpose(self, transpose_idx: List[int]) -> 'MixedState':
        self.system_seq = transpose_idx + [x for x in self._system_seq if x not in transpose_idx]
        transpose_dim = math.prod([self.system_dim[x] for x in transpose_idx])
        
        state = self.clone()
        state._data = utils.linalg._transpose_1(state._data, transpose_dim)
        return state

    def normalize(self) -> None:
        self._data = torch.div(self._data, utils.linalg._trace(self._data, -2, -1).view([-1, 1, 1]))

    def clone(self) -> 'MixedState':
        dim = self.dim
        data = self._data.view(self.batch_dim + [dim, dim]).clone()
        return MixedState(data, self.system_dim, self.system_seq, [prob.clone() for prob in self._prob])

    @DefaultSimulator.system_seq.setter
    def system_seq(self, target_seq: List[int]) -> None:
        if target_seq == self._system_seq:
            return
        
        perm_map = utils.linalg._perm_of_list(self._system_seq, target_seq)
        current_system_dim = [self._sys_dim[x] for x in self._system_seq]

        self._data = utils.linalg._permute_dm(self._data, perm_map, current_system_dim).contiguous()
        self._system_seq = target_seq
        
    def _index_select(self, new_indices: torch.Tensor) -> None:
        self._data = self._data.index_select(-2, new_indices).index_select(-1, new_indices)
    
    def _evolve(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        dim, _shape = self.dim, self._squeeze_shape
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim = unitary.shape[-1]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        data = self._data.view(_shape + [applied_dim, (dim ** 2) // applied_dim])
        data = torch.matmul(unitary, data)
        
        data = data.view(_shape + [dim, applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary.unsqueeze(-3).conj(), data).view([-1, dim, dim])
    
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, 
                     sys_idx: List[Union[int, List[int]]], on_batch: bool = True) -> None:
        dim, _shape = self.dim, self._squeeze_shape
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        ctrl_idx = sys_idx[0]
        ctrl_dim = np.prod([self.system_dim[idx] for idx in ctrl_idx])
        
        sys_idx = ctrl_idx + sys_idx[1:]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        applied_dim = unitary.shape[-1]
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        
        other_dim = dim // ctrl_dim        
        data = self._data.view(_shape + [ctrl_dim, applied_dim, (other_dim ** 2) * ctrl_dim // applied_dim]).clone()
        data[:, :, index] = torch.matmul(unitary, data[:, :, index].clone())
        
        data = data.view(_shape + [dim, ctrl_dim, applied_dim, other_dim // applied_dim])
        data[:, :, :, index] = torch.matmul(unitary.unsqueeze(-3).conj(), data[:, :, :, index].clone())
        
        self._data = data.view([-1, dim, dim])
    
    def _evolve_keep_dim(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        unitary_batch_dim = list(unitary.shape[:-2])
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = (self._batch_dim or unitary_batch_dim[:-1]) + unitary_batch_dim[-1:]
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim, num_unitary = unitary.shape[-1], int(np.prod(unitary_batch_dim[-1:]))
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [num_unitary, applied_dim, applied_dim])
        data = self._data.view(_shape + [1, applied_dim, (dim ** 2) // applied_dim])
        data = torch.matmul(unitary, data)
        
        data = data.view(_shape + [num_unitary, dim, applied_dim, dim // applied_dim])
        unitary = unitary.unsqueeze(-3)
        data = torch.matmul(unitary.conj(), data)
        self._data = data.view([-1, dim, dim])
    
    def _expec_val(self, obs: torch.Tensor, sys_idx: List[int]) -> torch.Tensor:
        dim = self.dim
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        applied_dim, num_obs = obs.shape[-1], obs.shape[-3]
        state = self._data.view([-1, 1, applied_dim, (dim ** 2) // applied_dim])
        state = torch.matmul(obs, state).view([-1, dim, dim])
    
        return utils.linalg._trace(state, -2, -1).view(self.batch_dim + [num_obs])
    
    def _expec_state(self, prob_idx: List[int]) -> 'MixedState':
        dim, num_prob = self.dim, len(self._prob)
        batch_prob_len = len(self._prob[-1].shape[:-num_prob])
        prob = self._joint_probability(prob_idx)
        
        states = self._data.view([-1] + self._prob_dim + [dim ** 2])
        prob = prob.view(list(prob.shape) + [1] * (self._prob[-1].dim() - prob.dim() + 1))
        prob_state = torch.mul(prob, states)

        sum_idx = [idx + 1 for idx in prob_idx]
        expectation = prob_state.sum(sum_idx)

        new_prob = []
        if len(prob_idx) != num_prob:
            new_prob = [prob.clone() for idx, prob in enumerate(self._prob) if idx not in prob_idx]
            expectation = expectation.view(self._batch_dim + list(new_prob[-1].shape[batch_prob_len:]) + [dim, dim])
        else:
            expectation = expectation.view(self._batch_dim + [dim, dim])
        return MixedState(expectation, self.system_dim, self.system_seq, new_prob)
    
    def _measure(self, measure_op: torch.Tensor, sys_idx: List[int]) -> Tuple[torch.Tensor, 'MixedState']:
        new_state = self.clone()
        new_state._evolve_keep_dim(measure_op, sys_idx)
        
        data = new_state._data
        measure_prob = utils.linalg._trace(data, -2, -1).real.view([-1, 1, 1])
        mask = measure_prob >= 1e-10
        collapsed_data = data / torch.where(mask, measure_prob, torch.ones_like(measure_prob))
        collapsed_data *= mask.to(collapsed_data.dtype)
        collapsed_data[torch.isnan(collapsed_data)] = 0
        
        measure_prob = measure_prob.view(new_state._batch_dim[:-1] + self._prob_dim + [-1])
        new_state._data = collapsed_data
        new_state._batch_dim = new_state._batch_dim[:-1]
        new_state._prob.append(measure_prob)
        return measure_prob, new_state
    
    def __kraus_transform(self, list_kraus: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Apply the Kraus operator to the state.
        
        Args:
            list_kraus: the Kraus operators.
            sys_idx: the system index list.
            on_batch: whether this operator evolves on batch axis. Defaults to True.
        
        """
        self._evolve_keep_dim(list_kraus, sys_idx)
        self._batch_dim, rank = self._batch_dim[:-1], self._batch_dim[-1]
        self._data = self._data.view([-1, rank, self.dim, self.dim]).sum(dim=-3)
    
    def __choi_transform(self, choi: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Apply the Choi operator to the state.
        
        Args:
            choi: the Choi operator.
            sys_idx: the system index list.
            on_batch: whether this operator evolves on batch axis. Defaults to True.
        
        """
        _shape = self._squeeze_shape
        if on_batch:
            self._batch_dim = self._batch_dim or list(choi.shape[:-2])
            transform_axis = [-1, 1]
        else:
            transform_axis = [1, -1]
        
        refer_sys_idx = [x for x in self._system_seq if x not in sys_idx]
        dim_refer = int(np.prod([self.system_dim[x] for x in refer_sys_idx]))
        dim_in = int(np.prod([self.system_dim[x] for x in sys_idx]))
        dim_out = choi.shape[-1] // dim_in
        
        self.system_seq = refer_sys_idx + sys_idx
        data = self._data.view(_shape + [dim_refer, dim_in, dim_refer, dim_in])
        choi = choi.view(transform_axis + [dim_in, dim_in * (dim_out ** 2)])
        
        data = torch.matmul(data.transpose(-1, -3).reshape(_shape + [(dim_refer ** 2) * dim_in, dim_in]), choi)
        data = torch.transpose(data.view(_shape + [dim_in, dim_refer, dim_out, dim_in, dim_out]), -3, -4)
        data = utils.linalg._trace(data, -2, -5)
        self._data = data.view([-1, dim_refer * dim_out, dim_refer * dim_out])
    
    def _transform(self, op: torch.Tensor, sys_idx: List[int], repr_type: str) -> None:
        self.__kraus_transform(op, sys_idx) if repr_type == 'kraus' else self.__choi_transform(op, sys_idx)
    
    def sqrt(self) -> torch.Tensor:
        return utils.linalg._sqrtm(self.density_matrix)
    
    def log(self) -> torch.Tensor:
        if self.rank < self.dim:
            warnings.warn(
                f"The matrix logarithm may not be accurate: expect rank {self.dim}, received {self.rank}", UserWarning)
        return utils.linalg._logm(self.density_matrix)


class PureState(DefaultSimulator):
    r"""The pure state class.

    Args:
        data: tensor array in vector representation for quantum pure state(s).
        sys_dim: a list of dimensions for each system.
        system_seq: the system order of this state. Defaults to be from 1 to n.
        probability: tensor array for state distributions. Defaults to be 1.

    Note:
        The data is stored in the vector-form with shape :math:`(-1, d)`

    """
    backend: str = 'default-pure'
    
    def __init__(self, data: torch.Tensor, sys_dim: List[int], 
                 system_seq: Optional[List[int]] = None,
                 probability: Optional[List[torch.Tensor]] = None):
        super().__init__(data, sys_dim, system_seq, probability)
        
        dim = self.dim
        self._data = data.reshape([-1, dim])
        
    def __getitem__(self, key: Union[int, slice]) -> 'PureState':
        assert self.batch_dim, \
            f"This state is not batched and hence cannot be indexed: received key {key}."
        return PureState(self.ket[key], self.system_dim, self.system_seq, 
                         [prob.clone() for prob in self._prob])
    
    def prob_select(self, outcome_idx: torch.Tensor, prob_idx: int = -1) -> 'PureState':
        num_prob = len(self._prob)
        if prob_idx > 0:
            prob_idx -= num_prob

        new_prob = []
        for idx, prob in enumerate(self._prob):
            if num_prob + prob_idx > idx:
                new_prob.append(prob.clone())
            else:
                new_prob.append(prob.index_select(dim=prob_idx, index=outcome_idx).squeeze(prob_idx))
        
        data_idx = prob_idx - 2
        data = self.ket.index_select(dim=data_idx, index=outcome_idx).squeeze(data_idx)
        return PureState(data, self.system_dim, self.system_seq, new_prob)
    
    def _expand(self, batch_dim: List[int]) -> 'PureState':
        if original_batch_dim := self.batch_dim:
            assert batch_dim[-len(original_batch_dim):] == original_batch_dim, \
                    f"Expand dimension mismatch: expected {original_batch_dim}, received {batch_dim[-len(original_batch_dim):]}."
        
        expand_state = self.clone()
        expand_state._data = self.ket.expand(batch_dim + [-1, -1]).reshape([-1, self.dim])
        return expand_state
    
    def add_probability(self, prob: torch.Tensor) -> None:
        num_previous_prob = len(self._prob)
        prob = prob.view([1] * num_previous_prob + [-1])

        data = self._data.view(self.batch_dim + [1, self.dim])
        self._data = data.repeat(self.batch_dim + [prob.shape[-1], 1]).view([-1, self.dim])
        self._prob.append(prob)
    
    @staticmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        # formalize the input data with [d], [1, d], [d, 1], [..., d, 1]
        if data.ndim == 1:
            data = data.view([-1, 1])
        else:
            data = data.view(list(data.shape[:-2]) + [-1, 1])

        if isinstance(sys_dim, int):
            num_sys = int(math.log(data.shape[-2], sys_dim))
            expected_dimension = sys_dim ** num_sys
        else:
            num_sys = len(sys_dim)
            expected_dimension = int(np.prod(sys_dim))
        
        assert data.shape[-2] == expected_dimension, \
            f"Input data shape {data.shape} does not match the input system dimension {sys_dim}."
        
        if eps:
            assert torch.all(is_vec := utils.check._is_state_vector(data, eps)), \
                f"(Some) data is not a state vector: asserted {is_vec.tolist()}"
        
        return num_sys
        
    @property
    def ket(self) -> torch.Tensor:
        self.reset_sequence()
        return self._data.view(self.batch_dim + [self.dim, 1]).clone()

    @property
    def density_matrix(self) -> torch.Tensor:
        ket = self.ket
        return torch.matmul(ket, ket.mH)
    
    def to_mixed(self) -> MixedState:
        r"""Convert the pure state to mixed state computation.
        """
        return MixedState(self.density_matrix, self.system_dim, self.system_seq, 
                           [prob.clone() for prob in self._prob])
    
    def _trace(self, sys_idx: List[int]) -> MixedState:
        return self.to_mixed()._trace(sys_idx)
    
    def _reset(self, reset_idx: List[int], replace_state: 'DefaultSimulator') -> 'MixedState':
        return self.to_mixed()._reset(reset_idx, replace_state)
    
    def _transpose(self, sys_idx: List[int]) -> MixedState:
        return self.to_mixed()._transpose(sys_idx)

    @property
    def rank(self) -> int:
        return 1
    
    def normalize(self) -> None:
        self._data = torch.div(self._data, torch.norm(self._data, dim=-1))

    def clone(self) -> 'PureState':
        data = self._data.view(self.batch_dim + [self.dim, 1]).clone()
        state = PureState(data, self.system_dim, self.system_seq, [prob.clone() for prob in self._prob])
        state._keep_dim = self._keep_dim
        return state

    @DefaultSimulator.system_seq.setter
    def system_seq(self, target_seq: List[int]) -> None:
        if target_seq == self._system_seq:
            return
        
        perm_map = utils.linalg._perm_of_list(self._system_seq, target_seq)
        current_system_dim = [self._sys_dim[x] for x in self._system_seq]
        
        self._data = utils.linalg._permute_sv(self._data, perm_map, current_system_dim).contiguous()
        self._system_seq = target_seq
        
    def _index_select(self, new_indices: torch.Tensor) -> None:
        self._data = self._data.index_select(dim=-1, index=new_indices)
        
    def _record_unitary(self, unitary: torch.Tensor, sys_idx: List[int]) -> None:
        r"""The function that records the input unitary for computing the overall 
        unitary matrix of the circuit.
        
        Note:
            This function is for calling `Circuit.matrix` only
        """
        if self._batch_dim == [self.dim]:
            self._evolve_keep_dim(unitary, sys_idx)
            return
        
        applied_dim, dim = unitary.shape[-1], self.dim
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        data = self._data.view([dim, -1, applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary, data).view([-1, dim])
        
    def _evolve(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        if self._keep_dim:
            self._record_unitary(unitary, sys_idx)
            return
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim = unitary.shape[-1]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        data = self._data.view(_shape + [applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary, data).view([-1, dim])
    
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, sys_idx: List[Union[int, List[int]]], on_batch: bool = True) -> None:
        ctrl_idx = sys_idx[0]
        ctrl_dim = np.prod([self.system_dim[idx] for idx in ctrl_idx])
        
        sys_idx = ctrl_idx + sys_idx[1:]
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        if self._keep_dim:
            proj = torch.zeros([ctrl_dim, ctrl_dim])
            proj[index, index] = 1
            unitary = torch.kron(proj, unitary) + torch.kron(torch.eye(ctrl_dim) - proj, torch.eye(unitary.shape[-1]).expand_as(unitary))
            self._record_unitary(unitary, sys_idx)
            return
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = self._batch_dim or list(unitary.shape[:-2])
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim = unitary.shape[-1]
        unitary = unitary.view(evolve_axis + [applied_dim, applied_dim])
        
        other_dim = dim // ctrl_dim
        data = self._data.view(_shape + [ctrl_dim, applied_dim, other_dim // applied_dim]).clone()
        data[:, :, index] = torch.matmul(unitary, data[:, :, index].clone())
        
        self._data = data.view([-1, dim])
        
    def _evolve_keep_dim(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        unitary_batch_dim = list(unitary.shape[:-2])
        dim, _shape = self.dim, self._squeeze_shape
        
        if on_batch:
            self._batch_dim = (self._batch_dim or unitary_batch_dim[:-1]) + unitary_batch_dim[-1:]
            evolve_axis = [-1, 1]
        else:
            evolve_axis = [1, -1]
        
        applied_dim, num_unitary = unitary.shape[-1], int(np.prod(unitary_batch_dim[-1:]))
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        unitary = unitary.view(evolve_axis + [num_unitary, applied_dim, applied_dim])
        data = self._data.view(_shape + [1, applied_dim, dim // applied_dim])
        self._data = torch.matmul(unitary, data).view([-1, dim])
    
    def _expec_val(self, obs: torch.Tensor, sys_idx: List[int]) -> torch.Tensor:
        dim = self.dim
        self.system_seq = sys_idx + [x for x in self._system_seq if x not in sys_idx]
        
        applied_dim, num_obs = obs.shape[-1], obs.shape[-3]
        state = self._data.view([-1, 1, applied_dim, dim // applied_dim])
        measured_state = torch.matmul(obs, state).view([-1, num_obs, dim])

        return torch.linalg.vecdot(self._data.unsqueeze(-2), measured_state).view(self.batch_dim + [num_obs])
    
    def _expec_state(self, prob_idx: List[int]) -> 'PureState':
        dim, num_prob = self.dim, len(self._prob)
        batch_prob_len = len(self._prob[-1].shape[:-num_prob])
        prob = self._joint_probability(prob_idx)
        
        states = self._data.view([-1] + self._prob_dim + [dim])
        prob = prob.view(list(prob.shape) + [1] * (self._prob[-1].dim() - prob.dim() + 1))
        prob_state = torch.mul(prob, states)
        
        sum_idx = [idx + 1 for idx in prob_idx]
        expectation = prob_state.sum(sum_idx)

        new_prob = []
        if len(prob_idx) != num_prob:
            new_prob = [prob.clone() for idx, prob in enumerate(self._prob) if idx not in prob_idx]
            expectation = expectation.view(self._batch_dim + list(new_prob[-1].shape[batch_prob_len:]) + [dim, 1])
        else:
            expectation = expectation.view(self._batch_dim + [dim, 1])
        return PureState(expectation, self.system_dim, self.system_seq, new_prob)
    
    def _measure(self, measure_op: torch.Tensor, sys_idx: List[int]) -> Tuple[torch.Tensor, 'PureState']:
        new_state = self.clone()
        new_state._evolve_keep_dim(measure_op, sys_idx)
        
        data = new_state._data.view([-1, self.dim, 1])
        measure_prob = (data.mH @ data).real
        mask = measure_prob >= 1e-10
        collapsed_data = data / torch.sqrt(torch.where(mask, measure_prob, torch.ones_like(measure_prob)))
        collapsed_data *= mask.to(collapsed_data.dtype)
        collapsed_data[torch.isnan(collapsed_data)] = 0
        
        measure_prob = measure_prob.view(new_state._batch_dim[:-1] + self._prob_dim + [-1])
        new_state._data = collapsed_data.view([-1, self.dim])
        new_state._batch_dim = new_state._batch_dim[:-1]
        new_state._prob.append(measure_prob)
        return measure_prob, new_state

    def _transform(self, *args) -> None:
        raise NotImplementedError(
            "The state vector backend does not internally support the channel conversion. \
                Please call the 'transform' function directly instead of '_transform'.")
    
    def transform(self, op: torch.Tensor, sys_idx: List[int] = None, repr_type: str = 'kraus') -> MixedState:
        return self.to_mixed().transform(op, sys_idx, repr_type)
    
    def product_trace(self, trace_state: 'PureState', trace_idx: List[int]) -> 'PureState':
        r"""Partial trace over this state, when this state is a product state
        """
        remain_seq, remain_system_dim, trace_system_dim = [], [], []
        for x in self._system_seq:
            if x in trace_idx:
                trace_system_dim.append(self.system_dim[x])
            else:
                remain_seq.append(x)
                remain_system_dim.append(self.system_dim[x])

        assert trace_state.system_dim == trace_system_dim, \
                f"Traced state does not match with the target trace system: received {trace_state.system_dim}, expect {trace_system_dim}."
        if trace_state.batch_dim:
            assert trace_state.batch_dim == self.batch_dim[-len(trace_state.batch_dim):], \
                f"# of traced states mismatch: received {trace_state.batch_dim}, expect {self.batch_dim[-len(trace_state.batch_dim):]}."
            num_trace = np.prod(trace_state.batch_dim)
        else:
            num_trace = 1

        trace_state = trace_state.ket.squeeze(-1)
        self.system_seq = trace_idx + remain_seq

        data = self._data.view([-1] + [num_trace, self.dim])
        data = utils.linalg._ptrace_1(data, trace_state)
        data = data.view(self.batch_dim + [-1, 1])
        
        # convert remaining sequence
        value_to_index = {value: index for index, value in enumerate(sorted(remain_seq))}
        remain_seq = [value_to_index[i] for i in remain_seq]
        return PureState(data, remain_system_dim, remain_seq, self._prob)
    
    def sqrt(self) -> torch.Tensor:
        return self.density_matrix
    
    def log(self) -> torch.Tensor:
        raise torch.zeros_like(self.density_matrix)




# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Source file for base class that simulates quantum states.
"""


import warnings
from abc import abstractmethod
from typing import Iterable, List, Optional, Tuple, Type, Union

import numpy as np
import torch

from ... import Hamiltonian, utils
from .base import State


class StateSimulator(State):
    r"""The abstract base class for simulating quantum states in QuAIRKit.

    Args:
        data: tensor array for quantum state(s).
        sys_dim: a list of dimensions for each system.
        system_seq: the system order of this state. Defaults to be from 0 to n - 1.
        probability: list of state probability distributions. Defaults to be 1.

    Note:
        This class is designed to be inherited by specific state simulator classes.
        Such class should simulate quantum computing with batch, qudit and probabilities.

    """
    def __init__(self, data: torch.Tensor, sys_dim: List[int],
                 system_seq: Optional[List[int]],
                 probability: Optional[List[torch.Tensor]]) -> None:
        super().__init__(sys_dim)

        self._prob = [] if probability is None else probability
        non_batch_len = 2 + len(self._prob)
        self._batch_dim = list(data.shape[:-non_batch_len])

        self._data = data.contiguous().to(dtype=self.dtype, device=self.device)
        self._system_seq = list(range(len(sys_dim))) if system_seq is None else system_seq

        self._keep_dim = False # the flag to switch on the unitary matrix recording.

    @abstractmethod
    def __getitem__(self, key: Union[int, slice]) -> 'StateSimulator':
        r"""Indexing of the State class
        """

    @abstractmethod
    def prob_select(self, outcome_idx: torch.Tensor, prob_idx: int = -1) -> 'StateSimulator':
        r"""Indexing probability outcome from the State batch along the given dimension.

        Args:
            outcome_idx: the 1-D tensor containing the outcomes to index
            prob_idx: the int that indexing which probability distribution. Defaults to be the last distribution.

        Returns:
            States that are selected by the specific probability outcomes.
        
        """
    
    @abstractmethod
    def add_probability(self, prob: torch.Tensor) -> None:
        r"""Add one probability distribution to the state.
        
        Args:
            prob: the probability distribution to be added, which should be a 1-D tensor.
        
        """

    @abstractmethod
    def expand_as(self, other: 'StateSimulator') -> 'StateSimulator':
        r"""Expand this tensor to the same size as other.

        Args:
            other: the state to be expanded to.

        Note:
            See torch.Tensor.expand_as() for more information about expand logic.

        """

    def __copy__(self) -> 'StateSimulator':
        return self.clone()

    @abstractmethod
    def __str__(self) -> str:
        r"""String representation of the state.
        """
        
    def __repr__(self) -> None:
        return self.__str__()

    def __matmul__(self, other: 'StateSimulator') -> torch.Tensor:
        r"""Matrix product between the density matrix representation of two states

        Args:
            other: a quantum state

        Returns:
            the product of these two states

        """
        return torch.matmul(self.density_matrix, other.density_matrix)

    def _check_op_dim(self, op: torch.Tensor, sys_idx: List[int]) -> None:
        applied_dim = op.shape[-1]
        assert applied_dim == int(np.prod((list_dim := [self._sys_dim[i] for i in sys_idx]))), \
            f"The operator's dimension {applied_dim} does not match the dimensions of its acting systems {list_dim}"

    @abstractmethod
    def kron(self, other: 'StateSimulator') -> 'StateSimulator':
        r"""Take the tensor product with other state.
        """

    @staticmethod
    @abstractmethod
    def check(data: torch.Tensor, sys_dim: Union[int, List[int]], eps: Optional[float] = 1e-4) -> int:
        r"""Assert whether the input data is valid for the specific State class.

        Args:
            data: tensor array for quantum state(s).
            sys_dim: (list of) dimension(s) of the systems, can be a list of integers or an integer.
            eps: the tolerance for the numerical check. Can be None means the check is overridden.

        Returns:
            the number of systems.

        """

    @abstractmethod
    def _joint_probability(self, prob_idx: List[int]) -> torch.Tensor:
        r"""The joint probability distribution of these states' occurrences

        Args:
            prob_idx: the indices of probability distributions

        """

    @property
    @abstractmethod
    def probability(self) -> torch.Tensor:
        r"""The probability distribution(s) of these states' occurrences
        """

    @property
    @abstractmethod
    def _prob_dim(self) -> List[int]:
        r"""Current probability dimensions
        """

    @property
    @abstractmethod
    def batch_dim(self) -> List[int]:
        r"""The batch dimension of this state
        """

    def numel(self) -> int:
        r"""The number of elements in this data
        """
        return int(np.prod(self.batch_dim))

    @property
    def _squeeze_shape(self) -> List[int]:
        r"""The squeezed shape of this state batch
        """
        return [-1, int(np.prod(self._prob_dim))]

    @property
    @abstractmethod
    def shape(self) -> torch.Size:
        r"""The recognize shape of this state
        """

    @property
    @abstractmethod
    def ket(self) -> torch.Tensor:
        r"""The ket form of this state

        Note:
            If the state is pure, the ket form is the statevector representation.
            If the state is mixed, the ket form is the vectorization of its density matrix representation.

        """

    @property
    def bra(self) -> torch.Tensor:
        r"""Dagger of the ket form.
        """
        return self.ket.mH

    @property
    @abstractmethod
    def density_matrix(self) -> torch.Tensor:
        r"""The density matrix representation of this state.
        """

    @property
    def vec(self) -> torch.Tensor:
        r"""Vectorization of the state: :math:`\textrm{vec}(|i \rangle\langle j|)=|j, i \rangle`
        """
        data = self.density_matrix.view([1, -1, self.dim, self.dim])
        
        newshape = [data.shape[0], data.shape[1], -1, 1]
        rev_src = tuple(range(data.ndim - 1, -1, -1))
        rev_dst = tuple(reversed(newshape))
        data = data.permute(rev_src).reshape(rev_dst).permute(tuple(range(len(newshape) - 1, -1, -1)))
        return data.view(self.batch_dim + [-1, 1])

    @abstractmethod
    def _trace(self, trace_idx: List[int]) -> 'StateSimulator':
        r"""Partial trace of the state

        Args:
            trace_idx: the subsystem indices to be traced out.

        Returns:
            the partial trace of the state as a new state.

        """
        
    @abstractmethod
    def _reset(self, reset_idx: List[int], replace_state: 'StateSimulator') -> 'StateSimulator':
        r"""Reset the state to a new state.

        Args:
            reset_idx: the subsystem indices to be reset.
            replace_state: the state to replace the quantum state.
            
        Note:
            reset_idx should not include all systems, otherwise call `StateSimulator.reset`.
        
        """

    @abstractmethod
    def _transpose(self, transpose_idx: List[int]) -> 'StateSimulator':
        r"""Partial transpose of the state

        Args:
            transpose_idx: the subsystem indices to be transposed.

        Returns:
            the transposed state as a new state.

        """

    @property
    @abstractmethod
    def rank(self) -> List[int]:
        r"""The rank of the state.
        """

    @abstractmethod
    def normalize(self) -> None:
        r"""Normalize this state to the correct format
        """

    @abstractmethod
    def numpy(self) -> np.ndarray:
        r"""Get the data in numpy.

        Returns:
            The numpy array of the data for the quantum state.
        """

    @abstractmethod
    def clone(self) -> 'StateSimulator':
        r"""Return a copy of the quantum state.
        """

    @abstractmethod
    def to(self, dtype: str = None, device: str = None) -> 'StateSimulator':
        r"""Change the property of the data tensor, and return a copy of this State

        Args:
            dtype: the new data type of the state.
            device: the new device of the state.

        """

    @abstractmethod
    def _index_select(self, new_indices: torch.Tensor) -> None:
        r"""Select the entries stored in this state.

        Args:
            new_indices: the indices of the entries to be selected.
        """

    def index_select(self, new_indices: Iterable[int]) -> 'StateSimulator':
        r"""Select the entries stored in this state.

        Args:
            new_indices: the new indices of the entries to be selected.

        Returns:
            a new state with selected indices.

        """
        if not isinstance(new_indices, torch.Tensor):
            new_indices = torch.tensor(new_indices, dtype=torch.long, device=self.device)

        assert new_indices.numel() == self.dim, \
            f"The number of new indices {new_indices.numel()} does not match the state dimension {self.dim}."

        new_state = self.clone()
        new_state._index_select(new_indices)
        return new_state

    @abstractmethod
    def permute(self, target_seq: List[int]) -> 'StateSimulator':
        r"""Permute the systems order of the state.

        Args:
            target_seq: the target systems order.

        """

    @abstractmethod
    def _evolve(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Evolve this state with unitary operators.

        Args:
            unitary: the unitary operator.
            sys_idx: the system indices to be acted on.
            on_batch: whether this unitary operator evolves on batch axis. Defaults to True.

        Note:
            The difference between `State.evolve` and `State._evolve` is that the former returns a new state 
            by calling an extra clone.

            Support the following transformation
            - (), () -> ()
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error

        """

    @abstractmethod
    def _evolve_keep_dim(self, unitary: torch.Tensor, sys_idx: List[int], on_batch: bool = True) -> None:
        r"""Evolve this state with unitary operators, while record the size of unitary batch

        Args:
            unitary: the unitary operator.
            sys_idx: the system indices to be acted on.
            on_batch: whether this unitary operator evolves on batch axis. Defaults to True.

        Note:
            Support the following transformation
            - (), () -> (1)
            - (m), () -> (m)
            - (), (n) -> (n, 1)
            - (m), (n) -> (n, m)
            - (n, m), (n) -> (n, m)
            - (l, m), (n) -> error
        """

    @abstractmethod
    def _evolve_ctrl(self, unitary: torch.Tensor, index: int, sys_idx: List[Union[int, List[int]]]) -> None:
        r"""Evolve this state with unitary operators controlled by a computational state

        Args:
            unitary: the unitary operator.
            index: the index of the computational state that activates the unitary.
            sys_idx: indices of the systems on which the whole unitary is applied. The first element in the list is 
                a list that gives the control system, while the remaining elements are int that give the applied system.

        Note:
            Support the following transformation
            - (), () -> ()
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error

        """

    @abstractmethod
    def _transform(self, op: torch.Tensor, sys_idx: List[int], repr_type: str, on_batch: bool = True) -> None:
        r"""Apply a general linear operator to the state.

        Args:
            op: the input operator.
            sys_idx: the subsystem indices to be applied.
            repr_type: the representation type of input operator, can be 'kraus' or 'choi'.
            on_batch: whether this operator evolves on batch axis. Defaults to True.

        Note:
            The difference between `State.transform` and `State._transform` is that the former returns a new state 
            by calling an extra clone.

            Support the following transformation
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error

        """

    @abstractmethod
    def _expec_val(self, obs: torch.Tensor, sys_idx: List[int]) -> torch.Tensor:
        r"""The expectation value of observables.

        Args:
            obs: the (list of) input observable.
            sys_idx: the system indices to be measured.

        Returns:
            the expectation value of the input observable for the quantum state.

        Note:
            Support the following transformation
            - (), (n) -> (n)
            - (n), () -> (n)
            - (n), (n) -> (n)
            - (m), (n) -> error

        """

    @abstractmethod
    def _expec_state(self, prob_idx: List[int]) -> 'StateSimulator':
        r"""The expectation with respect to the specific probability distribution(s) of states

        Args:
            prob_idx: indices of probability distributions. Defaults to all distributions.
        """

    @abstractmethod
    def _measure(self, measure_op: torch.Tensor, sys_idx: List[int]) -> Tuple[torch.Tensor, 'StateSimulator']:
        r"""Measure the quantum state with the measured operators.

        Args:
            measure_op: the measurement operators, where the first dimension is the number of operators.

        Returns:
            The probability and collapsed states of each measurement result.

        Note:
            Support the following transformation
            - (r), () -> (r)
            - (m, r), () -> (m, r)
            - (r), (n) -> (n, r)
            - (n, r), (n) -> (n, r)
            - (m, r), (n) -> error

        """

    def evolve(self, unitary: torch.Tensor, sys_idx: Optional[Union[int, List[int]]] = None) -> 'StateSimulator':
        r"""Evolve this state with unitary operators.

        Args:
            unitary: the unitary operator.
            sys_idx: the system indices to be acted on. Defaults to all systems.

        Returns:
            the evolved state.

        Note:
            Evolve support the broadcast rule.

        """
        sys_idx = self._check_sys_idx(sys_idx)
        self._check_op_dim(unitary, sys_idx)

        new_state = self.clone()
        new_state._evolve(unitary, sys_idx)
        return new_state

    def transform(self, op: torch.Tensor, sys_idx: Optional[Union[int, List[int]]] = None,
                  repr_type: str = 'kraus') -> 'StateSimulator':
        r"""Apply a general linear operator to the state.

        Args:
            op: the input operator.
            sys_idx: the qubit indices to be applied. Defaults to all systems.
            repr_type: the representation type of input operator. Defaults to 'kraus'.

        Returns:
            the transformed state.

        """
        # TODO add assertion for the dimension of input channels
        sys_idx = self._check_sys_idx(sys_idx)

        # TODO add support for batched operators
        repr_type = repr_type.lower()
        if repr_type == 'kraus':
            self._check_op_dim(op, sys_idx)
            if len(op.shape) == 2:
                op = op.unsqueeze(0)
            elif len(op.shape) > 3:
                raise NotImplementedError(
                    'consider the batched Kraus operators in the upcoming future')

        elif repr_type == 'choi':
            if len(op.shape) > 2:
                raise NotImplementedError(
                    'consider the batched Choi operators in the upcoming future')
        else:
            raise ValueError(
                f"Unsupported representation type {repr_type}: expected 'kraus' or 'choi'.")

        new_state = self.clone()
        new_state._transform(op, sys_idx, repr_type)
        return new_state

    def product_trace(self, trace_state: 'StateSimulator', trace_idx: List[int]) -> 'StateSimulator':
        r"""Partial trace over this state, when this state is a product state

        Args:
            trace_state: the state for the subsystem to be traced out.
            trace_idx: the subsystem indices to be traced out.

        Note:
            This function only works when the state is a product state represented by PureState

        """
        return self.trace(trace_idx)

    def expec_val(self, hamiltonian: Hamiltonian, shots: Optional[int] = None, decompose: bool = False) -> torch.Tensor:
        r"""The expectation value of the observable with respect to the quantum state.

        Args:
            hamiltonian: Input observable.
            shots: The number of shots to measure the observable. Should not be used in simulation mode.
            decompose: If decompose is ``True``, it will return the expectation value of each term.

        Returns:
            The expectation value (per term) of the input observable for the quantum state.

        Note:
            currently only run in qubit case.
            
        Raises:
            NotImplementedError: If `shots` is specified, since simulators do not support shot-based computations.
            NotImplementedError: If the state is not a qubit state.
            AssertionError: If the number of qubits in the Hamiltonian is greater than the number of state qubits.

        """
        if shots:
            raise NotImplementedError(
                "The shots argument is not supported in simulators. "
                "Please use the subclass of `StateOperator` for shot-based computations.")
        
        if not self.are_qubits():
            raise NotImplementedError(
                f"Currently only support qubit computation in Hamiltonian tasks: received {self._sys_dim}")

        assert hamiltonian.n_qubits <= self.num_qubits, \
            f"The number of qubits in the Hamiltonian {hamiltonian.n_qubits} is greater the number of state qubits {self.num_qubits}"

        num_terms, list_qubits_idx = hamiltonian.n_terms, hamiltonian.sites
        list_coef, list_matrices = hamiltonian.coefficients, hamiltonian.pauli_words_matrix

        expec_val_each_term = []
        for i in range(num_terms):
            qubits_idx = list_qubits_idx[i]
            if qubits_idx == ['']:
                expec_val_each_term.append(list_coef[i] * self.trace())
                continue
            matrix = list_matrices[i]

            #TODO this assertion should be done by Hamiltonian, not State
            assert 2 ** len(qubits_idx) == matrix.shape[0], \
                f"The qubit index {qubits_idx} does not match the matrix dimension {matrix.shape[0]}"

            expec_val = self._expec_val(matrix.unsqueeze(0), qubits_idx).squeeze(-1).real * list_coef[i]
            expec_val_each_term.append(expec_val)

        expec_val_each_term = torch.stack(expec_val_each_term)
        return expec_val_each_term if decompose else torch.sum(expec_val_each_term, dim=0)

    def expec_state(self, prob_idx: Optional[Union[int, List[int]]] = None) -> 'StateSimulator':
        r"""The expectation with respect to the specific probability distribution(s) of states

        Args:
            prob_idx: indices of probability distributions. Defaults to all distributions.

        Returns:
            The expected State obtained from the taken probability distributions.

        """
        if self.probability.numel() == 0:
            return self.clone()
        num_prob = len(self._prob)

        if prob_idx is None:
            prob_idx = list(range(num_prob))
        elif isinstance(prob_idx, int):
            prob_idx = [prob_idx]
        else:
            prob_idx = [num_prob + idx if idx < 0 else idx for idx in sorted(prob_idx)]
        return self._expec_state(prob_idx)

    def measure(self, system_idx: Optional[Union[int, List[int]]] = None, shots: Optional[int] = None,
                measure_op: Optional[torch.Tensor] = None, 
                is_povm: bool = False, keep_state: bool = False) -> Union[torch.Tensor, Tuple[torch.Tensor, 'StateSimulator']]:
        r"""Measure the quantum state

        Args:
            system_idx: the system indices to be measured. Defaults to all systems.
            shots: the number of shots to measure. Should not be used in simulation mode.
            measure_op: measurement operator. Defaults to the computational basis.
            is_povm: whether the measurement is a POVM.
            keep_state: whether to keep the state after measurement.

        Returns:
            The probability of each measurement result. The corresponding collapsed state will also be returned if
            `is_povm` is False or `keep_state` is True.

        Raises:
            NotImplementedError: If `shots` is specified, since simulators do not support shot-based computations.
            UserWarning: `is_povm` and `keep_state` cannot be both True,
            since a general POVM does not distinguish states.

        """
        if shots:
            raise NotImplementedError(
                "The shots argument is not supported in simulators. "
                "Please use the subclass of `StateOperator` for shot-based computations.")
        
        system_idx = self._check_sys_idx(system_idx)

        if measure_op is None:
            dim = int(np.prod([self._sys_dim[i] for i in system_idx]))
            identity = torch.eye(dim, dtype=self.dtype, device=self.device).unsqueeze(-1)
            measure_op = identity @ identity.mH
        else:
            self._check_op_dim(measure_op, system_idx)

        if is_povm:
            if keep_state:
                raise ValueError(
                    "`is_povm` and `keep_state` cannot be both True, " +
                    "since a general POVM does not distinguish states.")
            return self._expec_val(measure_op, system_idx)

        prob, collapsed_state = self._measure(measure_op, system_idx)
        return (prob, collapsed_state) if keep_state else prob

    @abstractmethod
    def sqrt(self) -> torch.Tensor:
        r"""Matrix square root of the state.
        """

    @abstractmethod
    def log(self) -> torch.Tensor:
        r"""Matrix logarithm of the state.
        """

    def trace(self, trace_idx: Optional[Union[List[int], int]] = None) -> Union[torch.Tensor, 'StateSimulator']:
        r"""(Partial) trace of the state

        Args:
            trace_idx: the subsystem indices to be traced out. Defaults to all systems.

        Returns:
            the trace of the state as a Tensor, or a new state if sys_idx is not None.
        """
        if trace_idx is None:
            return utils.linalg._trace(self.density_matrix, -2, -1)
        if isinstance(trace_idx, int):
            trace_idx = [trace_idx]
        assert max(trace_idx) < self.num_systems, \
            f"The trace index {trace_idx} should be smaller than number of systems {self.num_systems}"

        return self._trace(trace_idx)
    
    def reset(self, reset_idx: Union[List[int], int], replace_state: 'StateSimulator') -> 'StateSimulator':
        r"""Reset the state to a new state.

        Args:
            reset_idx: the subsystem indices to be reset.
            replace_state: the state to replace the quantum state.

        Returns:
            the new state after resetting.

        """
        if isinstance(reset_idx, int):
            reset_idx = [reset_idx]
        system_dim = [self.system_dim[i] for i in reset_idx]
        assert replace_state.system_dim == system_dim, \
            f"The system dimension of the replace state {replace_state.system_dim} does not match the reset system dimension {system_dim}."
        if len(reset_idx) == self.num_systems:
            warnings.warn("All systems will be reset: gradient break here.", UserWarning)
            return replace_state.clone()
        if replace_state.batch_dim:
            raise NotImplementedError(
                "A batched replace state is not supported")
        
        return self._reset(reset_idx, replace_state)

    def transpose(self, transpose_idx: Optional[Union[List[int], int]] = None) -> 'StateSimulator':
        r"""(Partial) transpose of the state

        Args:
            transpose_idx: the subsystem indices to be transposed. Defaults to all systems.

        Returns:
            the transpose of the state as a Tensor, or a new state if sys_idx is not None.
        """
        if transpose_idx is None:
            transpose_idx = list(range(self.num_systems))
        elif isinstance(transpose_idx, int):
            transpose_idx = [transpose_idx]
        assert max(transpose_idx) < self.num_systems, \
            f"The transpose index {transpose_idx} should be smaller than number of systems {self.num_systems}"

        return self._transpose(transpose_idx)



# !/usr/bin/env python3
# Copyright (c) 2024 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The library of functions in quantum information theory and quantum computing.
"""

from .check import *
from .linalg import *
from .qinfo import *

from .check import __all__ as check_fcn
from .linalg import __all__ as linalg_fcn
from .qinfo import __all__ as qinfo_fcn

__all__ = check_fcn + linalg_fcn + qinfo_fcn




# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Source file for basic operator.
"""


import copy
import uuid
from typing import Any, Dict, List, Optional, TypedDict, Union

import torch

from ..base import get_device, get_dtype
from .transpile import OpenQASM2Transpiler


class OperatorInfoType(TypedDict):
    r"""A dictionary storing basic information for Operator.

    Keys:
        - uid: Unique identifier for the operator.
        - name: Operator name.
        - type: Operator type, can be one of:
            - 'general': General operator.
            - 'gate': Quantum gate.
            - 'channel': Quantum channel.
            - 'locc': One-way local operation and classical communication.
        - tex: LaTeX representation (without '$').
        - system_idx: Physical system index or indices the operator acts upon. Shape:
            ``[# of operators, num_acted_system]``
        - num_ctrl_system: Number of indices that control the operator.
        - label: Labels for control systems, collapse events (measurement), or measurement outcomes.
        - param: Parameter tensor for parameterized operators. Shape:
            ``[1 or len(system_idx), batch_size, num_acted_param]``
        - param_sharing: Boolean indicating if parameters are shared. Defaults to ``False``.
        - matrix: Matrix representation (if type is 'gate').
        - api: Circuit API method that invokes the operator in a circuit.
        - permute: Index representation for permutation operators.
        - kwargs: Additional keyword arguments for circuit APIs.
        - plot_width: Plot width (optional, default: ``None``).
    
    Note:
        This class is used as a type hint for `OperatorInfo`, and should not be instantiated directly.
        It is designed to provide a structured way to store and access operator information in QuAIRKit.
    
    """

    uid: str
    name: str
    type: str
    tex: str
    system_idx: List[List[int]]
    num_ctrl_system: int
    label: Union[List[str], List[int], str]
    param: torch.Tensor
    param_sharing: bool
    matrix: torch.Tensor
    api: str
    permute: List[int]
    kwargs: Dict[str, Any]
    plot_width: Optional[float]  # TODO depreciated in the next version
    
    def __init__(self, *args, **kwargs) -> None:
        raise NotImplementedError(
            "OperatorInfoType is used as a type hint for OperatorInfo, and hence should not be instantiated.")
    
    @property
    def qasm(self) -> str:
        r"""Display in OpenQASM-like format
        """
        pass
    
    @property
    def qasm2(self) -> str:
        r"""Display in OpenQASM 2.0 format
        """
        pass     


class OperatorInfo(dict):
    __slots__ = ()
    __allowed_keys = {
        "uid",
        "name",
        "type",
        "tex",
        "system_idx",
        "num_ctrl_system",
        "label",
        "param",
        "param_sharing",
        "matrix",
        "api",
        "permute",
        "kwargs",
        "plot_width",  # TODO depreciated in the next version
    }

    def __init__(self, name: str = "unidentified", type: str = "general", **kwargs) -> None:
        uid = str(uuid.uuid4())
        super().__init__(uid=uid, name=name, type=type, **kwargs)

    def __setitem__(self, key, value):
        if key not in self.__allowed_keys:
            raise KeyError(
                f"Invalid key '{key}' for OperatorInfo, accept one of {self.__allowed_keys}."
            )

        elif key == "uid" and self.get("uid"):
            raise KeyError(
                "uid is automatically generated and should not be set manually."
            )

        elif key == "type":
            if value not in {"general", "gate", "channel", "locc"}:
                raise ValueError(
                    f"Invalid type '{value}' for OperatorInfo, accept one of "
                    "{'general', 'gate', 'channel', 'locc'}."
                )

        elif key == "system_idx":
            if not isinstance(value, List) or not all(isinstance(idx, List) for idx in value):
                raise TypeError(f"system_idx must be a list of lists: received {value}")
            if not all(isinstance(idx, int) for sublist in value for idx in sublist):
                raise TypeError(f"All indices in system_idx must be integers: received {value}")
            value = copy.deepcopy(value)

        elif key == "param":
            if not isinstance(value, torch.Tensor):
                raise TypeError(f"param must be a torch.Tensor: received type {type(value)}")
            if value.ndim != 3:
                raise ValueError(f"param tensor must be 3-dimensional: received shape {value.shape}")
            if (original_param := self.get('param')) is not None and original_param.shape != value.shape:
                raise ValueError(
                    f"Changing param shape from {original_param.shape} to {value.shape} is not allowed."
                )

        elif key == "matrix":
            if self['type'] != "gate":
                raise ValueError(f"matrix is only valid for gate type operators, not {self['type']}.")
            if not isinstance(value, torch.Tensor):
                raise TypeError("matrix must be a torch.Tensor")

        super().__setitem__(key, value)
        
    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        
        new_instance = OperatorInfo(name=self['name'], type=self['type'])
        memo[id(self)] = new_instance
        
        for key in self.keys():
            if key == "uid":
                continue
            value = self[key]
            
            if isinstance(value, torch.Tensor):
                # FIXME Sometimes requires_grad is not set correctly without manually setting requires_grad_(True)
                new_instance[key] = value.clone().requires_grad_(True) if value.requires_grad else value.clone()
            else:
                new_instance[key] = copy.deepcopy(value, memo)
                
        return new_instance

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    @property
    def qasm(self) -> str:
        r"""Display in OpenQASM-like format
        """
        return OpenQASM2Transpiler.format_qasm(self['name'], self['system_idx'], self.get('param', None))

    @property
    def qasm2(self) -> str:
        r"""Display in OpenQASM 2.0 format
        """
        return OpenQASM2Transpiler.transpile(self['name'], self['system_idx'], self.get('param', None))


class Operator(torch.nn.Module):
    r"""The basic class to implement the operation in QuAIRKit.

    Args:
        dtype: The data type of the operator.
            Defaults to ``None``, which means to use the default data type.

    """

    def __init__(self, dtype: Optional[str] = None):
        if dtype is None:
            self.dtype = get_dtype()
            super().__init__()
        else:
            self.dtype = dtype
            super().__init__(dtype=dtype)
        self.device = get_device()

        self._info: OperatorInfoType = OperatorInfo()

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if isinstance(value, Operator) and value.dtype is None:
            value.dtype = self.dtype

    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this operator."""
        return copy.deepcopy(self._info)

    def to(self, dtype: Optional[str] = None) -> None:
        if dtype is not None:
            self.dtype = dtype
            for sub_layer in self.children():
                sub_layer.dtype = dtype

    def forward(self, *inputs, **kwargs) -> None:
        raise NotImplementedError

    def dagger(self) -> None:
        raise NotImplementedError(
            "General operations are not physically reversible. Please check whether the circuit "
            "contains non-unitary operations such as general channels or measurement.")



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


r"""
The source file of the LOCCNet class.
"""

import warnings
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import torch
from torch.nn import ModuleDict, ModuleList

from quairkit import Circuit
from quairkit.core import StateSimulator, get_dtype, intrinsic
from quairkit.core.intrinsic import _alias
from quairkit.database import bell_state, std_basis, zero_state
from quairkit.operator import OneWayLOCC, Oracle, ParamOracle
from quairkit.qinfo import nkron

_LogicalIndex = Tuple[str, int]
_PartyInfo = Dict[str, Union[int, Dict[str, int]]]


class OneWayLOCCNet(torch.nn.Module):
    r"""Network for general one-way LOCC (Local Operations and Classical Communication) protocol.
    Party that is measured will be traced out, and the party that performs the action will be the only one left.
    
    Args:
        party_info: a dictionary containing the information of the parties involved in the LOCC protocol.
        
    Note:
        Information of each party in `party_info` should be one of the following format:
        - {'Alice': 2} (Alice has 2 qubits)
        - {'Alice': {'num_systems': 2, 'system_dim': 3}} (Alice has 2 qutrits)
        - {'Alice': {'num_systems': 2, 'system_dim': [2, 3]}} (Alice has a qubit and a qutrit)
    
    """
    def __init__(self, party_info: _PartyInfo) -> None:
        super().__init__()
        self._cir_map = ModuleDict()
        self._state_map: Dict[Tuple[int, ...], StateSimulator] = {}
        self._list_locc = ModuleList()

        current_idx = 0
        for name, info in party_info.items():
            if isinstance(info, int):
                num_systems = info
                system_dim = [2] * num_systems
            else:
                num_systems = info['num_systems']
                dim_info = info.get('system_dim', 2)
                system_dim = dim_info if isinstance(dim_info, List) else [dim_info] * num_systems
            
            idx_map = list(range(current_idx, current_idx + num_systems))
            self._cir_map[name] = Circuit(num_systems, system_dim, physical_idx=idx_map)
            current_idx += num_systems
        
    def __getitem__(self, key: str) -> Circuit:
        return self._cir_map[key]
    
    def __setitem__(self, key: str, cir: Circuit) -> None:
        r"""Set a party's circuit in the LOCC protocol.
        
        Args:
            key: The name of the party.
            cir: The circuit of the party.
        
        """
        if not isinstance(cir, Circuit):
            raise TypeError("Value must be an instance of Circuit")

        if key in self._cir_map:
            existing_cir = self._cir_map[key]
            if cir.num_systems != existing_cir.num_systems or cir.system_dim != existing_cir.system_dim:
                raise ValueError(
                    f"Circuit mismatch for '{key}': expected systems {existing_cir.num_systems}, dimensions {existing_cir.system_dim}; "
                    f"got systems {cir.num_systems}, dimensions {cir.system_dim}"
                )
        self._cir_map[key] = cir
        
    def __delitem__(self, key: str) -> None:
        raise NotImplementedError(
            "Deleting parties from the LOCC protocol is not supported.")
    
    def __len__(self) -> int:
        return len(self._cir_map)
    
    def __repr__(self):
        return repr(self._cir_map)
    
    def __str__(self):
        return str(self._cir_map)
    
    def keys(self) -> Iterable[str]:
        r"""Get the names of the parties in the LOCC protocol.
        """
        return self._cir_map.keys()
    
    def values(self) -> Iterable[Circuit]:
        r"""Get the circuits in the LOCC protocol.
        """
        return self._cir_map.values()
    
    def items(self) -> Iterable[Tuple[str, Circuit]]:
        return self._cir_map.items()
        
    def __check_party(self, *party_names: str) -> None:
        r"""Check if the party is registered in the LOCC protocol.
        """
        available_parties = set(self.keys())
        for name in party_names:
            if name not in available_parties:
                raise KeyError(f"Party '{name}' is not registered. Available parties: {available_parties}")
    
    @property
    def party_info(self) -> _PartyInfo:
        r"""Get the information of the parties in the LOCC protocol.
        """
        return {
            name: {'num_systems': circuit.num_systems, 'system_dim': circuit.system_dim}
            for name, circuit in self.items()
        }
    
    @property
    def physical_circuit(self) -> Circuit:
        r"""Get the complete physical circuit of the LOCC protocol.
        """
        system_dim = sum((cir.system_dim for cir in self.values()), [])
        combined_circuit = Circuit(system_dim=system_dim)
        for cir in self.values():
            combined_circuit += cir
            
        for op in self._list_locc:
            combined_circuit.append(op)
        return combined_circuit
    
    @_alias({"system_idx": "qubits_idx"})
    def set_init_state(self, system_idx: Union[Union[List[_LogicalIndex], _LogicalIndex]], state: Optional[StateSimulator] = None) -> None:
        r"""Set the initial (Bell) state of the LOCC protocol.
        
        Args:
            system_idx: a list of logical indices of where this state is an input
            state: the initial state of the LOCC protocol. Defaults to the (generalized) Bell state pair
        
        """
        if isinstance(system_idx, Tuple):
            system_idx = [system_idx]
        
        physical_indices, indexed_system_dim = [], []
        for name, idx in system_idx:
            self.__check_party(name)
            circuit = self._cir_map[name]
            physical_indices.append(circuit.system_idx[idx])
            indexed_system_dim.append(circuit.system_dim[idx])

        indices_tuple = tuple(physical_indices)
        if indices_tuple in self._state_map:
            raise ValueError(f"Initial state for indices {indices_tuple} already set.")

        if state is None:
            if len(physical_indices) != 2:
                raise ValueError("Bell state initialization requires exactly two systems.")
            state = bell_state(2, system_dim=indexed_system_dim)
        elif state.system_dim != indexed_system_dim:
            raise ValueError(f"State dimension mismatch: expected {indexed_system_dim}, got {state.system_dim}")

        state.reset_sequence()
        self._state_map[indices_tuple] = state
        
    def __prepare_init_state(self) -> StateSimulator:
        r"""Prepare the physical initial state for the LOCC protocol.
        """
        num_total_systems = sum(cir.num_systems for cir in self.values())
        total_system_dim = sum((cir.system_dim for cir in self.values()), [])

        settled_indices = [idx for indices in self._state_map for idx in indices]
        unsettled_indices = sorted(set(range(num_total_systems)) - set(settled_indices))

        states = [self._state_map[indices] for indices in self._state_map]
        if unsettled_indices:
            unsettled_dims = [total_system_dim[idx] for idx in unsettled_indices]
            states.append(zero_state(len(unsettled_indices), system_dim=unsettled_dims))

        combined_state = nkron(*states)
        combined_state._system_seq = settled_indices + unsettled_indices
        return combined_state
    
    @_alias({"system_idx": "qubits_idx"})
    def locc(self, local_unitary: torch.Tensor, system_idx: List[Union[List[_LogicalIndex], _LogicalIndex]],
             label: str = 'M', latex_name: str = 'O') -> None:
        r"""Set a (non-parameterized) one-way LOCC protocol to the network.
        
        Args:
            local_unitary: The local unitary operation.
            system_idx: Systems on which the protocol is applied. The first element indicates the measure system.
            label: Label for measurement. Defaults to 'M'.
            latex_name: LaTeX name for the applied operator. Defaults to 'O'.
        
        """
        if isinstance(system_idx[0], Tuple):
            system_idx[0] = [system_idx[0]]
        measure_idx, apply_idx = [], []
        measure_dim, apply_dim = [], []
        
        for name, idx in system_idx[0]:
            cir = self[name]
            measure_idx.append(cir.system_idx[idx])
            measure_dim.append(cir.system_dim[idx])
            
        for idx, op in enumerate(self._list_locc):
            settled_idx = op.measure.system_idx
            assert set(measure_idx).isdisjoint(set(settled_idx)), \
                f"Measurement system {measure_idx} has been used in the {idx}-th LOCC protocols {settled_idx}"
            
        for name, idx in system_idx[1:]:
            cir = self[name]
            apply_idx.append(cir.system_idx[idx])
            apply_dim.append(cir.system_dim[idx])
        
        gate = Oracle(local_unitary, apply_idx, apply_dim)
        self._list_locc.append(OneWayLOCC(gate, measure_idx, measure_dim, label=label, latex_name=latex_name))
       
    @_alias({"system_idx": "qubits_idx"})
    def param_locc(self, generator: Callable[[torch.Tensor], torch.Tensor], num_acted_param: int, 
                   system_idx: List[Union[List[_LogicalIndex], _LogicalIndex]], param: Union[torch.Tensor, float] = None, 
                   label: str = 'M', latex_name: str = 'U', support_batch: bool = True) -> None:
        r"""Add a one-way LOCC protocol comprised of unitary operations, where the applied unitary is parameterized.

        Args:
            generator: Function to generate the oracle.
            num_acted_param: Number of parameters required for a single application.
            system_idx: Systems on which the protocol is applied. The first element indicates the measure system.
            param: Input parameters for the gate. Defaults to None.
            label: Label for measurement. Defaults to 'M'.
            latex_name: LaTeX name for the applied operator. Defaults to 'U'.
            support_batch: Whether generator supports batched input.
        
        Note:
            If the generator does not support batched input, you need to set `support_batch` to `False`.

        """
        if isinstance(system_idx[0], Tuple):
            system_idx[0] = [system_idx[0]]
        measure_idx, apply_idx = [], []
        measure_dim, apply_dim = [], []
        
        for name, idx in system_idx[0]:
            cir = self[name]
            measure_idx.append(cir.system_idx[idx])
            measure_dim.append(cir.system_dim[idx])
            
        for idx, op in enumerate(self._list_locc):
            settled_idx = op.measure.system_idx
            assert set(measure_idx).isdisjoint(set(settled_idx)), \
                f"Measurement system {measure_idx} has been used in the {idx}-th LOCC protocols {settled_idx}"
            
        for name, idx in system_idx[1:]:
            cir = self[name]
            apply_idx.append(cir.system_idx[idx])
            apply_dim.append(cir.system_dim[idx])
        
        if param is None:
            float_dtype = intrinsic._get_float_dtype(get_dtype())
            expect_shape = intrinsic._format_param_shape([apply_idx], num_acted_param, param_sharing=False, batch_size=np.prod(measure_dim))
            param = torch.nn.Parameter(torch.rand(expect_shape, dtype=float_dtype) * 2 * np.pi)
        param_gate = ParamOracle(
            generator, apply_idx, param, num_acted_param, apply_dim, support_batch=support_batch)
        self._list_locc.append(OneWayLOCC(param_gate, measure_idx, measure_dim, label=label, latex_name=latex_name))
    
    def __call__(self) -> StateSimulator:
        return self.forward()
        
    def forward(self) -> StateSimulator:
        r"""Run the one-way LOCC protocol and return the final state.
        
        Returns:
            The final state of the LOCC protocol.
        
        """
        state = self.__prepare_init_state()
        for cir in self.values():
            state = cir(state)
        
        trace_idx = []
        trace_dim = []
        for op in self._list_locc:
            state = op(state)
            trace_idx.extend(op.measure.system_idx)
            trace_dim.append(np.prod(op.measure.system_dim))
        trace_state = std_basis(len(trace_idx), system_dim=[state.system_dim[idx] for idx in trace_idx])
        trace_state._batch_dim = trace_dim
        return state.product_trace(trace_state, trace_idx)



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Source file for transpilation between OperatorInfo and [OpenQASM 2.0](https://github.com/openqasm/openqasm/tree/OpenQASM2.x)
"""

from typing import Dict, Iterable, List, Optional, Union

import torch


class OpenQASM2Transpiler(object):
    r"""A class to handle the translation between OperatorInfo and OpenQASM 2.0 format.
    
    This class provides methods to convert operator information into OpenQASM 2.0 format
    and to handle equivalent translations for specific operators.
    """
    
    __general_case: Dict[str, str] = {
        'h': 'h',
        'x': 'x',
        'y': 'y',
        'z': 'z',
        's': 's',
        'sdg': 'sdg',
        't': 't',
        'tdg': 'tdg',
        'rx': 'rx',
        'ry': 'ry',
        'rz': 'rz',
        'p': 'u1',
        'u3': 'u3',
        'cx': 'cx',
        'cy': 'cy',
        'cz': 'cz',
        'ccx': 'ccx',
        'crz': 'crz',
        'cp': 'cu1',
    }
    r"""A dictionary of equivalent translation between OperatorInfo and OpenQASM 2.0.
    """
    
    __special_case: Dict[str, Dict] = {
        'cswap': {
            'num_qubits': 3,
            'num_param': 0,
            'qasm': lambda gate_idx: [
                f"cx q[{gate_idx[2]}], q[{gate_idx[1]}];",
                f"ccx q[{gate_idx[0]}], q[{gate_idx[1]}], q[{gate_idx[2]}];",
                f"cx q[{gate_idx[2]}], q[{gate_idx[1]}];"
            ]
        },
        'rxx': {
            'num_qubits': 2,
            'num_param': 1,
            'qasm': lambda gate_idx, params: [
                f"h q[{gate_idx[0]}];",
                f"h q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"rz({params[0]:.5f}) q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"h q[{gate_idx[0]}];",
                f"h q[{gate_idx[1]}];"
            ]
        },
        'ryy': {
            'num_qubits': 2,
            'num_param': 1,
            'qasm': lambda gate_idx, params: [
                f"rx(pi/2) q[{gate_idx[0]}];",
                f"rx(pi/2) q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"rz({params[0]:.5f}) q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"rx(-pi/2) q[{gate_idx[0]}];",
                f"rx(-pi/2) q[{gate_idx[1]}];"
            ]
        },
        'rzz': {
            'num_qubits': 2,
            'num_param': 1,
            'qasm': lambda gate_idx, params: [
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"rz({params[0]:.5f}) q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];"
            ]
        },
        'ms': {
            'num_qubits': 2,
            'num_param': 0,
            'qasm': lambda gate_idx: [
                f"h q[{gate_idx[0]}];",
                f"h q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"rz(-pi/2) q[{gate_idx[1]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"h q[{gate_idx[0]}];",
                f"h q[{gate_idx[1]}];"
            ]
        },
        'cu4': {
            'num_qubits': 2,
            'num_param': 4,
            'qasm': lambda gate_idx, params: [
                f"rz({params[3] * 2:.5f}) q[{gate_idx[0]}];",
                f"cu3({params[0]:.5f}, {params[1]:.5f}, {params[2]:.5f}) q[{gate_idx[0]}], q[{gate_idx[1]}];"
            ]
        },
        'swap': {
            'num_qubits': 2,
            'num_param': 0,
            'qasm': lambda gate_idx: [
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"cx q[{gate_idx[1]}], q[{gate_idx[0]}];",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];"
            ]
        },
        'crx': {
            'num_qubits': 2,
            'num_param': 1,
            'qasm': lambda gate_idx, params: [
                f"h q[{gate_idx[1]}];",
                f"crz({params[0]:.5f}) q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"h q[{gate_idx[1]}];"
            ]
        },
        'cry': {
            'num_qubits': 2,
            'num_param': 1,
            'qasm': lambda gate_idx, params: [
                f"ry({params[0] / 2:.5f}) q[{gate_idx[1]}]",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"ry({-params[0] / 2:.5f}) q[{gate_idx[1]}]",
                f"cx q[{gate_idx[0]}], q[{gate_idx[1]}];",
                f"ry({params[0] / 2:.5f}) q[{gate_idx[1]}]",
            ]
        },
    }
    r"""A dictionary of special cases for translation between OperatorInfo and OpenQASM 2.0.
    """
    
    @staticmethod
    def format_qasm(name: str, system_idx: List[List[int]], param: Optional[torch.Tensor]) -> str:
        r"""Translate OperatorInfo to OpenQASM-like format.

        Args:
            name: The name of the operator.
            system_idx: The indices of the systems the operator acts upon.
            param: The parameters of the operator, if any.

        Returns:
            The OpenQASM-like representation of the operator.
        
        """
        if param is not None:
            param = torch.remainder(param, 2 * torch.pi)
            param = _tensor_to_float(param.squeeze(1))
        
        gate_qasm = []
        for i, idx_group in enumerate(system_idx):
            param_str = ''
            if param is not None:
                param_str = '(' + ", ".join(["{:.5f}".format(p) for p in param[i]]) + ')'
            
            indices = ", ".join([f"q[{idx}]" for idx in idx_group])
            gate_qasm.append(f"{name}{param_str} {indices};")

        return '\n'.join(gate_qasm)
    
    @staticmethod
    def transpile(name: str, system_idx: List[List[int]], param: Optional[torch.Tensor]) -> str:
        r"""Translate OperatorInfo to OpenQASM 2.0 format.

        Args:
            name: The name of the operator.
            system_idx: The indices of the systems the operator acts upon.
            param: The parameters of the operator, if any.

        Returns:
            The OpenQASM 2.0 representation of the operator.
        """
        if name in OpenQASM2Transpiler.__general_case:
            return OpenQASM2Transpiler.format_qasm(OpenQASM2Transpiler.__general_case[name], system_idx, param)
        
        if translation := OpenQASM2Transpiler.__special_case.get(name):
            num_qubits, num_param, qasm = translation['num_qubits'], translation['num_param'], translation['qasm']
            
            if param is not None:
                if param.shape[1] != 1:
                    raise ValueError(f"Expected param to be not batched, got shape {param.shape}")
                assert param.shape[-1] == num_param, \
                    f"Expected {num_param} parameters, got {param.shape[-1]} for {name}."
                param = torch.remainder(param, 2 * torch.pi)
                param = _tensor_to_float(param.squeeze(1))

            gates_qasm = []
            for index, gate_idx in enumerate(system_idx):
                if len(gate_idx) != num_qubits:
                    raise ValueError(f"{name} requires {num_qubits} qubits, got {len(gate_idx)}")

                if param is None:
                    gates_qasm.extend(qasm(gate_idx))
                else:
                    gates_qasm.extend(qasm(gate_idx, param[index]))
            return "\n".join(gates_qasm)
        
        raise NotImplementedError(f"Operator {name} not supported in OpenQASM 2.0")


def _tensor_to_float(arr: torch.Tensor) -> Union[int, Iterable[float]]:
    r"""Convert a tensor to a nested list of float.
    """
    flat_list = list(arr.detach().cpu().numpy().reshape(-1))
    shape = arr.shape
    current = flat_list

    # Group elements into nested lists based on shape (from last dimension to first)
    for dim in reversed(shape):
        current = [current[i:i + dim] for i in range(0, len(current), dim)]

    # Extract the rebuilt nested list structure
    return current[0]



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
from abc import ABC, abstractmethod
from typing import List, Optional, Union

import numpy as np
import torch

from ... import Hamiltonian, base

__all__ = ['State']


class State(ABC):
    r"""The abstract base class for all quantum state backends in QuAIRKit.
    
    Args:
        sys_dim: a list of dimensions for each system.
    
    """
    backend: Optional[str] = None
    
    def __init__(self, sys_dim: List[int], 
                 dtype: Optional[torch.dtype] = None, device: Optional[torch.device] = None) -> None:
        assert all(d >= 2 for d in sys_dim), \
            f"All systems must have dimension >= 2: received {sys_dim}"
        
        self._dtype = dtype or base.get_dtype()
        self._device = device or base.get_device()
        
        self._sys_dim = sys_dim.copy()

    @property
    def dtype(self) -> torch.dtype:
        r"""The data type of this state
        """
        return self._dtype

    @property
    def device(self) -> torch.device:
        r"""The device of this state
        """
        return self._device

    @property
    def system_dim(self) -> List[int]:
        r"""The list of dimensions for each system
        """
        return self._sys_dim.copy()

    @system_dim.setter
    def system_dim(self, sys_dim: List[int]) -> None:
        assert int(np.prod(sys_dim)) == self.dim, \
            f"The input system dim {sys_dim} does not match the original state dim {self.dim}."
        self._sys_dim = sys_dim.copy()

    @property
    def dim(self) -> int:
        r"""The dimension of this state
        """
        return math.prod(self.system_dim)

    def are_qubits(self) -> bool:
        r"""Whether all systems are qubits
        """
        return all(x == 2 for x in self._sys_dim)

    def are_qutrits(self) -> bool:
        r"""Whether all systems are qutrits
        """
        return all(x == 3 for x in self._sys_dim)

    @property
    def num_systems(self) -> int:
        r"""The number of systems
        """
        return len(self._sys_dim)

    @property
    def num_qubits(self) -> int:
        r"""The number of qubits of this state, when all systems are qubits
        """
        return sum(x == 2 for x in self._sys_dim)

    def _check_sys_idx(self, sys_idx: Optional[Union[int, List[int]]]) -> List[int]:
        r"""Check the system dimension of the input system indices

        Args:
            sys_idx: the system indices to be checked.

        Returns:
            the list of system dimensions.

        """
        if sys_idx is None:
            sys_idx = list(range(self.num_systems))
        elif isinstance(sys_idx, int):
            sys_idx = [sys_idx]
        else:
            sys_idx = list(sys_idx)
            assert (
                all(isinstance(x, int) for x in sys_idx)
                and min(sys_idx) >= 0
                and max(sys_idx) < self.num_systems
            ), f"The input system indices {sys_idx} should be a list of integers in range [0, {self.num_systems - 1}]"
        return sys_idx
    
    @abstractmethod
    def clone(self) -> 'State':
        r"""Return a copy of the quantum state.
        """
        
    @abstractmethod
    def measure(self, system_idx: Optional[Union[int, List[int]]], 
                shots: Optional[int], measure_op: Union[str, Optional[torch.Tensor]]) -> torch.Tensor:
        r"""Measure the quantum state with the given measurement operator.
        
        Args:
            system_idx: the system indices to be measured.
            shots: the number of measurement shots.
            measure_op: the measurement operator.
            
        Returns:
            A tensor containing the measurement results.
        
        """
        
    @abstractmethod
    def expec_val(self, hamiltonian: Hamiltonian, shots: Optional[int], decompose: bool) -> torch.Tensor:
        r"""The expectation value of the observable with respect to the quantum state.

        Args:
            hamiltonian: Input observable.
            shots: the total number of shots to measure.
            decompose: If decompose is ``True``, it will return the expectation value of each term.

        Returns:
            The measured expectation value (per term) of the input observable for the quantum state.
        
        """




# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The source file of the class for the special quantum operator.
"""

import copy
import warnings
from typing import Iterable, List, Optional, Union

import numpy as np
import torch

from ..core import Operator, OperatorInfoType, StateSimulator, utils
from ..core.intrinsic import _alias, _digit_to_int, _int_to_digit
from . import Channel, Gate


class ResetState(Operator):
    r"""The class to reset the quantum state.
    
    Args:
        system_idx: list of systems to be reset.
        acted_system_dim: dimension of systems that this gate acts on.
        replace_state: the state to replace the quantum state.
        state_label: LaTeX label of the reset state, used for printing.
    """
    @_alias({"system_idx": "qubits_idx"})
    def __init__(self, system_idx: Union[int, Iterable[int]], acted_system_dim: List[int],
                 replace_state: StateSimulator, state_label: str) -> None:
        super().__init__()
        self.system_idx = [[system_idx]] if isinstance(system_idx, int) else [list(system_idx)]
        self.system_dim = acted_system_dim
        self.replace_state = replace_state
        
        _info = {"name": "reset",
                 "system_idx": self.system_idx,
                 "type": "channel",
                 "tex": state_label}
        self._info.update(_info)
        
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this operation
        """
        info = super().info
        info.update({
            'system_idx': copy.deepcopy(self.system_idx),
        })
        return info
    
    def forward(self, state: StateSimulator) -> StateSimulator:
        r"""Perform the reset operation on the input quantum state.
        
        Args:
            state: The input quantum state to be reset.
            
        Returns:
            The reset quantum state.
        """
        return state.reset(self.system_idx[0], self.replace_state)


class Collapse(Operator):
    r"""The class to compute the collapse of the quantum state.

    Args:
        system_idx: list of systems to be collapsed.
        acted_system_dim: dimension of systems that this gate acts on.
        desired_result: The desired result you want to collapse. Defaults to ``None`` meaning preserving all results, 
            and activate probabilistic computation.
        if_print: whether print the information about the collapsed state. Defaults to ``False``.
        measure_op: The measurement operators of the measurement. Defaults to the computational measure.

    """
    @_alias({"system_idx": "qubits_idx"})
    def __init__(self, system_idx: Union[int, Iterable[int]], acted_system_dim: List[int], 
                 desired_result: Union[int, str] = None, if_print: bool = False, measure_op: Optional[torch.Tensor] = None):
        super().__init__()
        self.measure_op = []

        system_idx = [system_idx] if isinstance(system_idx, int) else list(system_idx)
        if measure_op is None:
            dim = int(np.prod(acted_system_dim))
            identity = torch.eye(dim, dtype=self.dtype, device=self.device).unsqueeze(-1)
            measure_op = identity @ identity.mH
        else:
            assert len(measure_op) > 1, \
                    "The input measurement op should be a list of measurement operators."
            assert utils.check.is_pvm(measure_op), \
                    "The input measurement op do not form a projection-valued measurement."

        if desired_result is None:
            digits_str = None
        else:
            if isinstance(desired_result, str):
                desired_result = _digit_to_int(desired_result, acted_system_dim)
            digits_str = _int_to_digit(desired_result, acted_system_dim)
            desired_result = torch.tensor(desired_result)
            measure_op = torch.index_select(measure_op, -3, desired_result)

        self.system_idx = [system_idx]
        self.system_dim = acted_system_dim
        self.measure_op = measure_op

        self.if_print = if_print
        self.digits_str = digits_str
        self.desired_result = desired_result


        _info = {"name": "measure",
                 "system_idx": self.system_idx,
                 "type": "channel",
                 "kwargs": {"basis": measure_op, "if_print": if_print}}
        if desired_result is not None:
            _info["label"] = digits_str
        self._info.update(_info)
        
    def __call__(self, state: StateSimulator) -> StateSimulator:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)
    
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this channel
        """
        info = super().info
        info.update({
            'system_idx': copy.deepcopy(self.system_idx),
        })
        return info
        
    def forward(self, state: StateSimulator) -> StateSimulator:
        r"""Compute the collapse of the input state.

        Args:
            state: The input state, which will be collapsed

        Returns:
            The collapsed quantum state.
        """
        prob_array, measured_state = state._measure(self.measure_op, self.system_idx[0])
        
        desired_result = self.desired_result
        if desired_result is None:
            return measured_state
        
        digits_str = self.digits_str
        state_str = '>'.join(f'|{d}' for d in digits_str) + '>'
        
        if torch.any(prob_array < 1e-10).item():
            warnings.warn(
                f"It is computationally infeasible for some states in systems {self.system_idx} "
                f"to collapse to state {state_str}")

        # whether print the collapsed result
        if self.if_print:
            prob = prob_array.mean().item()
            print(f"systems {self.system_idx} collapse to the state {state_str} with (average) probability {prob}")

        return measured_state


class OneWayLOCC(Operator):
    r"""A one-way LOCC protocol, where quantum measurement is modelled by a PVM and all channels are unitary channels.
    
    Args:
        gate: a batched Gate instance.
        measure_op: The measurement operators of the measurement. Defaults to the computational measure.
        label: name of the measured label. Defaults to ``'M'``.
        latex_name: latex name of the applied operator. Defaults to ``'O'``.
    
    """
    @_alias({'system_idx': 'qubits_idx'})
    def __init__(
        self, gate: Gate,
        measure_idx: List[int], measure_dim: Union[List[int], int],
        measure_op: Optional[torch.Tensor] = None, label: str = 'M', latex_name: str = 'O'
    ):
        super().__init__()
        num_measure_system = len(measure_idx)
        measure_dim = [measure_dim] * num_measure_system if isinstance(measure_dim, int) else measure_dim
        
        assert gate.matrix.ndim > 2, \
            f"The local operation must be batched, received shape {gate.matrix.shape}"
        assert (batch_dim := gate.matrix.shape[0]) == np.prod(measure_dim), \
            f"Batch dimension mismatch: expected {np.prod(measure_dim)}, received {batch_dim} unitaries"
        
        self.measure = Collapse(measure_idx, measure_dim, measure_op=measure_op)
        self.local_op = gate
        
        self._info.update({'name': "locc",
                           'system_idx': self.system_idx,
                           'tex': latex_name,
                           'label': label,
                           'type': "locc",
                           'num_ctrl_system': len(measure_idx),
                           'kwargs': {'measure_basis': measure_op,
                                      'label_name': label}})
        
    def __call__(self, state: StateSimulator) -> StateSimulator:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)
    
    @property
    def system_idx(self) -> List[List[int]]:
        r"""The system indices of the LOCC protocol.
        """
        return [self.measure.system_idx[0] + self.local_op.system_idx[0]]
    
    @system_idx.setter
    def system_idx(self, value: List[List[int]]):
        r"""Set the system indices of the LOCC protocol.
        """
        value, system_idx = value[0], self.system_idx[0]
        assert len(value) == len(system_idx), \
             f"Length of system_idx should be {len(system_idx)}, but got {len(value)}"
        num_measure_system = len(self.measure.system_idx[0])
        self.measure.system_idx = [value[:num_measure_system]]
        self.local_op.system_idx = [value[num_measure_system:]]
    
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this operation
        """
        info = super().info
        info.update({
            'system_idx': copy.deepcopy(self.system_idx),
        })
        return info
    
    def forward(self, state: StateSimulator) -> StateSimulator:
        r"""Compute the input state passing through the LOCC protocol.

        Args:
            state: The input state.

        Returns:
            The collapsed quantum state.
        
        """
        matrix, sys_idx = self.local_op.matrix, self.local_op.system_idx[0]
        
        measured_state = self.measure(state)
        
        dim = len(matrix.shape[:-2])
        matrix = matrix.repeat(measured_state._prob_dim[:-dim] + [1] * (dim + 2))
        measured_state._evolve(matrix, sys_idx, on_batch=False)
        return measured_state


class QuasiOperation(Operator):
    r"""A quantum protocol containing quasi-operations.
    
    Args:
        channel: a batched Channel instance. Currently, we only support the gate channel.
        prob: a 1D (quasi-)probability distribution for the channel argument.
        probability_param: whether the probability is parameterized. Defaults to ``False``.
    
    """
    def __init__(
        self, channel: Channel, prob: torch.Tensor, probability_param: bool = False
    ):
        super().__init__()
        assert isinstance(channel, Gate), \
            f"Currently we only support quasi simulations with gates, received {channel.type_repr} representation"
        
        if (not probability_param) and np.abs((s := prob.sum().item()) - 1) > 1e-4:
            warnings.warn(
                f"The quasi-probability distribution should sum to 1, but received sum {s}. ", UserWarning)
        
        matrix = channel.matrix
        assert matrix.ndim == 3, \
            f"The input gate should be a batched gate, received shape {matrix.shape}."
            
        num_outcomes = prob.numel() + 1 if probability_param else prob.numel() 
        assert num_outcomes == (batch_num := matrix.shape[0]), \
            f"The prob distribution does not match the operator #: expect of num {num_outcomes}, received {batch_num}."
        
        prob = prob.flatten().to(dtype=self.dtype, device=self.device)
        if probability_param:
            prob = torch.nn.Parameter(prob)
            self.register_parameter('_prob', prob)
        else:
            self._prob = prob
        self.quasi_op = channel
        self.probability_param = probability_param
        
    def __call__(self, state: StateSimulator) -> StateSimulator:
        r"""Same as forward of Neural Network
        """
        return self.forward(state)
    
    @property
    def probability(self) -> torch.Tensor:
        r"""The (quasi-)probability distribution of the quasi-operation.
        """
        if self.probability_param:
            return torch.cat([self._prob, (1 - self._prob.sum()).view([-1])], dim=0)
        return self._prob
    
    @property
    def system_idx(self) -> List[int]:
        r"""The system indices of the quasi-operation.
        """
        return self.quasi_op.system_idx[0]
    
    @system_idx.setter
    def system_idx(self, value: List[int]):
        r"""Set the system indices of the quasi-operation.
        """
        assert len(value) == len(self.system_idx), \
             f"Length of system_idx should be {len(self.system_idx)}, but got {len(value)}"
        self.quasi_op.system_idx = [value]
    
    @property
    def info(self) -> OperatorInfoType:
        r"""Information of this operation
        """
        return self.quasi_op.info
    
    def forward(self, state: StateSimulator) -> StateSimulator:
        r"""Compute the input state passing through the quasi-operation.
        
        Args:
            state: The input state.
        
        Returns:
            The collapsed quantum state.
        """
        state = state.clone()
        state.add_probability(self.probability)
        
        matrix = self.quasi_op.matrix
        matrix = matrix.repeat(state._prob_dim[:-1] + [1, 1, 1])
        state._evolve(matrix, self.system_idx, on_batch=False)
        return state



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The module that implements various backends of the state.
"""

from typing import Dict, Optional, Type, Union

from .base import State
from .default import DefaultSimulator, MixedState, PureState
from .operator import StateOperator
from .simulator import StateSimulator

r"""Backend classifier in the QuAIRKit.
""" 
BACKEND_LIST: Dict[str, Type[State]] = {DefaultSimulator.backend: DefaultSimulator}

r"""Default simulator for the state backend.
"""
DEFAULT_BACKEND: str = DefaultSimulator.backend


def set_backend(backend: Union[str, State]) -> None:
    r"""Set the backend implementation of QuAIRKit state.

    Args:
        backend: The name of the backend.
    """
    global BACKEND_LIST
    list_backend = list(BACKEND_LIST.keys())

    if isinstance(backend, str):
        assert backend in BACKEND_LIST, \
            f"Backend '{backend}' is not founded in the backend list: {list_backend}."

        global DEFAULT_BACKEND
        DEFAULT_BACKEND = backend
        return

    assert issubclass(
        backend, (StateOperator, StateSimulator)
    ), f"New backend should be a subclass of `StateOperator` or 'StateSimulator', but got {type(backend)}"
    
    backend_str = backend.backend
    assert backend_str is not None, \
        f"New backend should have a non-trivial `backend` string attribute, but got {backend_str}."
    if backend_str in list_backend:
        assert backend.__name__ == (search_backend := BACKEND_LIST[backend_str].__name__), \
            (f"'{backend_str}' is already registered as {search_backend} in the backend list, "
             f"which is not the same as the new one {backend.__name__}.")
    
    BACKEND_LIST[backend_str] = backend
    DEFAULT_BACKEND = backend_str


def get_backend(backend: Optional[str] = None) -> Union[Type[StateOperator], Type[StateSimulator]]:
    r"""Get the current backend of QuAIRKit.

    Args:
        backend: The name of the backend.
    
    Returns:
        The state initializer of currently used backend.
    """
    backend = backend or DEFAULT_BACKEND
    assert backend in BACKEND_LIST, \
            f"Backend '{backend}' is not founded in the backend list: {list(BACKEND_LIST.keys())}."
    return BACKEND_LIST[DEFAULT_BACKEND]



# !/usr/bin/env python3
# Copyright (c) 2025 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Source file for base class that emulates quantum states.
"""


import copy
import itertools
import math
from abc import abstractmethod
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch

from ... import (Hamiltonian, OperatorInfo, OperatorInfoType, get_float_dtype,
                 utils)
from .base import State


def _convert_basis(qubits_idx: List[int], pauli_str: str) -> Tuple[List[OperatorInfoType], List[int]]:
    r"""Process Pauli string to get operators applied to the state before z-measurement.

    Args:
        qubits_idx: the indices of the qubits to be measured.
        pauli_str: the Pauli basis to measure, can be a char or a string composed of 'i', 'x', 'y', or 'z'.

    Returns:
        List of operators added to the circuit and the corresponding system indices.

    """
    obs_vals = torch.ones([1])
    dict_recover_op = {}
    for i, char in enumerate(pauli_str.lower()):
        idx = qubits_idx[i]
        if char == "x":
            obs = utils.matrix._z()
            dict_recover_op['h'] = dict_recover_op.get('h', []) + [[idx]]
        elif char == "y":
            obs = utils.matrix._z()
            dict_recover_op['sdg'] = dict_recover_op.get('sdg', []) + [[idx]]
            dict_recover_op['h'] = dict_recover_op.get('h', []) + [[idx]]
        elif char == "z":
            obs = utils.matrix._z()
        else:
            obs = utils.matrix._eye(2)
        
        obs_vals = torch.kron(obs_vals, obs.diag())
    obs_vals = obs_vals.real.int().tolist()
    
    list_recover_op = []
    if system_idx := dict_recover_op.get('sdg'):
        list_recover_op.append(OperatorInfo(name='sdg', system_idx=system_idx))
    if system_idx := dict_recover_op.get('h'):
        list_recover_op.append(OperatorInfo(name='h', system_idx=system_idx))
    return list_recover_op, obs_vals


class StateOperator(State):
    r"""The abstract base class for operating quantum state in QuAIRKit, via emulators or quantum devices.

    Args:
        list_operator: the operators of the circuit that applies to the state. Defaults to an empty list.
        sys_dim: a list of dimensions for each system.

    Note:
        This class is designed to be inherited by specific state emulator classes.
        Such class should emulate quantum computing via OpenQASM 2.0.

    """
    def __init__(self, list_operator: Optional[List[OperatorInfoType]], sys_dim: List[int]) -> None:
        super().__init__(sys_dim)

        self._data: List[OperatorInfoType] = copy.deepcopy(list_operator or [])

    @abstractmethod
    def _execute(self, qasm: str, shots: int) -> Dict[str, int]:
        r"""Execute a quantum circuit with the current operators and return the measurement results.
        
        Args:
            qasm: the QASM 2.0 string that represents the quantum circuit.
            shots: the number of shots to measure.
            
        Returns:
            A dictionary storing the number of each measurement outcomes.
            The keys are the outcome digits in binary string, and the values are the number of occurrences.
        """
    
    def _multi_execute(self, list_qasm: List[str], shots: int) -> Dict[str, List[int]]:
        r"""Execute multiple quantum circuits with the current (batched) operators and return the measurement results.
        
        Args:
            qasm: a list of QASM 2.0 strings that represent the quantum circuits.
            shots: the number of shots to measure for each circuit.
            
        Returns:
            A dictionary storing the number of each measurement outcomes.
            The keys are the outcome digits in binary string, and the values are the number of occurrences for each circuit.
        
        """
        assert isinstance(list_qasm, List) and all(isinstance(qasm, str) for qasm in list_qasm), \
            f"Expected a list of QASM strings, received {type(list_qasm)}."

        combined_result: Dict[str, List[int]] = {}
        for idx, qasm in enumerate(list_qasm):
            result = self._execute(qasm, shots)
            if idx == 0:
                for outcome in result.keys():
                    combined_result[outcome] = []
            
            for outcome in combined_result:
                combined_result[outcome].append(result.get(outcome, 0))
        return combined_result
    
    def __str__(self) -> str:
        return '\n'.join([op.qasm for op in self._data])
    
    @property
    def param(self) -> List[torch.Tensor]:
        r"""The parameters stored in the operators in StateOperator.
        """
        return [op['param'] for op in self._data if 'param' in op]

    def apply(self, list_operator: List[Union[OperatorInfoType, List[OperatorInfoType]]]) -> None:
        r"""Record a list of operators that will be applied to the device.

        Args:
            list_operator: the list of operators that applies to the state.

        """
        flat_list = []
        for op in list_operator:
            if isinstance(op, List):
                flat_list.extend(op)
            else:
                flat_list.append(op)
        self._data.extend(flat_list)

    @property
    def qasm2(self) -> str:
        r"""The QASM 2.0 string that represents the circuit stored in this state emulator.
        """
        assert self.are_qubits(), \
            "The StateOperator only supports qubit systems. Please use the simulator backend such as default simulator."

        qasm = [op.qasm2 for op in self._data]

        header = 'OPENQASM 2.0;\ninclude "qelib1.inc";'
        qreg = f"qreg q[{self.num_qubits}];\n"
        return '\n'.join([header, qreg] + qasm)
    
    def _measure_qasm2(self, qubits_idx: List[int]) -> str:
        r"""Generate the QASM 2.0 string that includes the measurement operation.
        
        Args:
            qubits_idx: the indices of the qubits to be measured.
            
        Returns:
            A QASM 2.0 string that includes the measurement operation for the specified qubits.
        """
        base_qasm = self.qasm2
        num_measure = len(qubits_idx)
        qasm_with_measure = f"{base_qasm}\n\ncreg c[{num_measure}];\n"
        if num_measure == self.num_qubits:
            qasm_with_measure += "measure q -> c;\n"
        else:
            for i, idx in enumerate(qubits_idx):
                qasm_with_measure += f"measure q[{idx}] -> c[{i}];\n"
        return qasm_with_measure
    
    @State.system_dim.setter
    def system_dim(self, sys_dim: List[int]) -> None:
        raise NotImplementedError(
            "A state operator should not change its system dimension after initialization.")
    
    @abstractmethod
    def clone(self) -> 'StateOperator':
        r"""Return a copy of the quantum state.
        """

    def _measure(self, qubits_idx: List[int], shots: int, pauli_str: str) -> Dict[str, Tuple[int, int]]:
        r"""Measure the quantum state

        Args:
            qubits_idx: the qubit indices to be measured. Defaults to all qubits.
            shots: the number of shots to measure, defaults to 1024.
            pauli_str: the Pauli basis to measure, can be a string composed of 'i', 'x', 'y', or 'z'.
            return_energy: whether the returned dictionary contains the energy (+-1) for each measurement outcome. Defaults to False.

        Returns:
            A dictionary storing the number (and energy) of each measurement outcome.
            The keys are the outcome digits in binary string, and the values are tuples of (number, energy).

        """
        list_recover_op, obs_vals = _convert_basis(qubits_idx, pauli_str)

        if len(list_recover_op):
            self._data.extend(list_recover_op)
            raw_result = self._execute(self._measure_qasm2(qubits_idx), shots)
            self._data = self._data[:-len(list_recover_op)]
        else:
            raw_result = self._execute(self._measure_qasm2(qubits_idx), shots)

        return {
            digits: (num_occur, obs_vals[int(digits, base=2)])
            for digits, num_occur in raw_result.items()
        }
    
    def _measure_probabilities(self, qubits_idx: List[int], shots: int, pauli_str: str) -> torch.Tensor:
        r"""Measure the quantum state and return the probabilities of each measurement outcome.
        
        Args:
            qubits_idx: the qubit indices to be measured. Defaults to all qubits.
            shots: the number of shots to measure, defaults to 1024.
            pauli_str: the Pauli basis to measure, can be a string composed of 'i', 'x', 'y', or 'z'.

        Returns:
            A dictionary storing the probability of each measurement outcome.
            The keys are the outcome digits in binary string, and the values are probabilities.

        """
        num_outcomes = 2 ** len(qubits_idx)
        probs = torch.zeros(num_outcomes, dtype=get_float_dtype())

        if shots == 0:
            return probs

        measurement_result = self._measure(qubits_idx, shots, pauli_str)
        for digits, (count, _) in measurement_result.items():
            idx = int(digits, 2)
            probs[idx] = count / shots
        return probs
    
    def measure(self, system_idx: Optional[Union[int, List[int]]] = None, 
                shots: Optional[int] = None, measure_op: Optional[str] = None) -> torch.Tensor:
        r"""Measure the quantum state
        
        Args:
            system_idx: the system indices to be measured. Defaults to all systems.
            shots: the number of shots to measure, defaults to 1024.
            measure_op: the measurement operator basis to measure. Here we restrict it to be Pauli, 
                which can be a char or a string composed of 'i', 'x', 'y', or 'z'. Defaults to 'z'.
            
        Returns:
            A tensor containing the measurement results, where each element corresponds to empirical probabilities of the measurement outcomes.
        """
        if shots is None:
            shots = 1024
        else:
            if shots < 0:
                raise ValueError(f"The number of shots should be a non-negative integer, received {shots}.")

            system_idx = self._check_sys_idx(system_idx)
            if shots == 0:
                return torch.zeros(2 ** len(system_idx), dtype=get_float_dtype())

        measure_op = 'z' if measure_op is None else measure_op.lower()
        if measure_op in {'x', 'y', 'z', 'i'}:
            measure_op *= len(system_idx)
        elif len(measure_op) != len(system_idx) or any(b not in {'x', 'y', 'z', 'i'} for b in measure_op):
            raise ValueError(
                f"Invalid format: received Pauli basis {measure_op} for the system indices {system_idx}.")

        return _compute_measure(self, system_idx, shots, measure_op, *self.param)
    
    def _expec_val_per_term(self, hamiltonian: Hamiltonian, list_shots: torch.Tensor) -> torch.Tensor:
        r"""The expectation value of each term in the observable with respect to the quantum state.

        Args:
            hamiltonian: Input observable.
            list_shots: the number of shots to measure for each term.

        Returns:
            The measured expectation value of each term in the input observable for the quantum state.

        """
        num_terms, list_qubits_idx = hamiltonian.n_terms, hamiltonian.sites
        list_coef, list_pauli_str = hamiltonian.coefficients, hamiltonian.pauli_words_r

        sum_per_term = []
        # TODO add parallel support, i.e. computing expectance values using multi_execute
        for i in range(num_terms):
            qubits_idx, coef, shots_per_term = list_qubits_idx[i], list_coef[i], list_shots[i].item()
            if qubits_idx == ['']:
                sum_per_term.append(coef * shots_per_term)
                continue
            
            measurement_result = self._measure(qubits_idx, shots_per_term, list_pauli_str[i])
            val = sum(num_occur * energy for num_occur, energy in measurement_result.values())
            
            sum_per_term.append(val * coef)
        
        sum_per_term, list_coef = torch.tensor(sum_per_term), torch.tensor(list_coef)
        
        expec_val_per_term = torch.zeros_like(sum_per_term)
        valid_indices = list_shots > 0
        expec_val_per_term[valid_indices] = sum_per_term[valid_indices] / list_shots[valid_indices]
        return expec_val_per_term
    
    def expec_val(self, hamiltonian: Hamiltonian, shots: Optional[int] = None, decompose: bool = False) -> torch.Tensor:
        r"""The expectation value of the observable with respect to the quantum state.

        Args:
            hamiltonian: Input observable.
            shots: the total number of shots to measure, defaults to 1024 times the Hamiltonian terms.
            decompose: If decompose is ``True``, it will return the expectation value of each term.

        Returns:
            The measured expectation value (per term) of the input observable for the quantum state.

        """
        if not self.are_qubits():
            raise NotImplementedError(
                f"Currently only support qubit computation in Hamiltonian tasks: received {self._sys_dim}")
        assert hamiltonian.n_qubits <= self.num_qubits, \
                f"The number of qubits in the Hamiltonian {hamiltonian.n_qubits} is greater than that of state qubits {self.num_qubits}"
        
        num_terms = hamiltonian.n_terms
        list_coef = torch.tensor(hamiltonian.coefficients)
        list_abs_coef = torch.abs(list_coef)
        
        shots = shots or 1024 * num_terms
        min_required_shots = 1 / list_abs_coef[list_abs_coef != 0].min() * 10
        if shots < min_required_shots:
            raise ValueError(
                f"Insufficient shots {shots} for Hamiltonian with coefficients {list_coef}.")
            
        pauli_weight = list_abs_coef / list_abs_coef.sum()
        pauli_sum = list_abs_coef.sum()
        list_shots = torch.from_numpy(np.random.multinomial(shots, pauli_weight))
        
        expec_val_per_term = _compute_expec_val_per_term(self, hamiltonian, list_shots, *self.param)
        if decompose:
            return expec_val_per_term
        
        sum_per_term = expec_val_per_term * list_shots
        factor_per_term = torch.zeros_like(sum_per_term)
        valid_indices = list_coef != 0
        factor_per_term[valid_indices] = pauli_sum / list_abs_coef[valid_indices]
        return (sum_per_term * factor_per_term).sum() / shots
        
    def _evaluate_for_grad(
        self, idx: int, param_idx: Tuple[int, int], shift: float,
        eval_func: Callable[['StateOperator'], torch.Tensor]
    ) -> torch.Tensor:
        r"""Helper function to evaluate quantum circuit at shifted parameter values.
        """
        i, j = param_idx
        with torch.no_grad():
            orig_val = float(self._data[idx]['param'][i, 0, j].item())
            self._data[idx]['param'][i, 0, j] = orig_val + shift
        result = eval_func(self)
        with torch.no_grad():
            self._data[idx]['param'][i, 0, j] = orig_val
        return result


class OperatorMeasure(torch.autograd.Function):
    @staticmethod
    def forward(state: StateOperator, qubits_idx: List[int], shots: int, 
                pauli_str: str, *params: torch.Tensor) -> torch.Tensor:
        return state._measure_probabilities(qubits_idx, shots, pauli_str)
    
    @staticmethod
    def setup_context(ctx, inputs, output):
        state, qubits_idx, shots, pauli_str, *params = inputs
        ctx.state = state.clone()
        
        def eval_func(input_state):
            return input_state._measure_probabilities(qubits_idx, shots, pauli_str)
        ctx.eval_func = eval_func
    
    # TODO add parallel support, i.e. computing shift values using multi_execute
    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        state: StateOperator = ctx.state.clone()
        eval_func = ctx.eval_func
        
        list_param_grad: List[Optional[torch.Tensor]] = []
        for idx, op_info in enumerate(state._data):
            if 'param' not in op_info:
                continue
                
            param_tensor = op_info['param']
            if not param_tensor.requires_grad:
                list_param_grad.append(None)
                continue
            num_op, num_acted_param = param_tensor.shape[0], param_tensor.shape[-1]
            param_grad = torch.zeros_like(param_tensor)
            
            shift_rule = get_shift_rule(op_info['name'])
            shifts = shift_rule[0]
            coeffs = shift_rule[1]
            num_shifts = len(shifts)
            
            for i, j in itertools.product(range(num_op), range(num_acted_param)):
                # Evaluate at shifted points
                results = [
                    state._evaluate_for_grad(idx, (i, j), shift, eval_func)
                    for shift in shifts
                ]
                expec_val_derivative = torch.stack([coeffs[k] * results[k] for k in range(num_shifts)]).sum(dim=0)
                
                # Apply chain rule: dL/dθ = (dL/dE)^T * (dE/dθ)
                param_grad[i, 0, j] = (expec_val_derivative * grad_output).sum()
            
            list_param_grad.append(param_grad)
        
        return None, None, None, None, *list_param_grad


class OperatorExpecVal(torch.autograd.Function):
    @staticmethod
    def forward(state: StateOperator, hamiltonian: Hamiltonian, 
                list_shots: torch.Tensor, *params: torch.Tensor) -> torch.Tensor:
        return state._expec_val_per_term(hamiltonian, list_shots)
    
    @staticmethod
    def setup_context(ctx, inputs, output):
        state, hamiltonian, list_shots, *params = inputs
        
        ctx.state = state.clone()
        def eval_func(input_state):
            return input_state._expec_val_per_term(hamiltonian, list_shots)
        ctx.eval_func = eval_func
    
    # TODO add parallel support, i.e. computing shift values using multi_execute
    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        state: StateOperator = ctx.state.clone()
        eval_func = ctx.eval_func
        
        list_param_grad: List[Optional[torch.Tensor]] = []
        for idx, op_info in enumerate(state._data):
            if 'param' not in op_info:
                continue
                
            param_tensor = op_info['param']
            if not param_tensor.requires_grad:
                list_param_grad.append(None)
                continue
            num_op, num_acted_param = param_tensor.shape[0], param_tensor.shape[-1]
            param_grad = torch.zeros_like(param_tensor)
            
            # Get shift rule for this gate
            shift_rule = get_shift_rule(op_info['name'])
            shifts = shift_rule[0]
            coeffs = shift_rule[1]
            num_shifts = len(shifts)
            
            for i, j in itertools.product(range(num_op), range(num_acted_param)):
                # Evaluate at shifted points
                results = [
                    state._evaluate_for_grad(idx, (i, j), shift, eval_func)
                    for shift in shifts
                ]
                expec_val_derivative = torch.stack([coeffs[k] * results[k] for k in range(num_shifts)]).sum(dim=0)
                
                param_grad[i, 0, j] = (expec_val_derivative * grad_output).sum()
            
            # Store gradient in the correct position
            list_param_grad.append(param_grad)
        
        return None, None, None, *list_param_grad


# Initialize shift rules, see [Quantum, 2022, 6: 677]
SHIFT_RULES = {
    'two_term': {
        'gates': ['rx', 'ry', 'rz', 'rxx', 'ryy', 'rzz', 'u3'],
        'shifts': [math.pi/2, -math.pi/2],
        'coeffs': [0.5, -0.5]
    },
    'four_term': {
        'gates': ['crx', 'cry', 'crz', 'cu4'],
        'shifts': [math.pi/4, -math.pi/4, 3*math.pi/4, -3*math.pi/4],
        'coeffs': [
            (math.sqrt(2) + 1) / (4 * math.sqrt(2)), 
            -(math.sqrt(2) + 1) / (4 * math.sqrt(2)), 
            -(math.sqrt(2) - 1) / (4 * math.sqrt(2)), 
            (math.sqrt(2) - 1) / (4 * math.sqrt(2))
        ]
    }
}
DELTA = 1e-2  # Default shift value for finite difference approximation


def get_shift_rule(gate_name: str) -> Optional[Tuple[List[float], List[float]]]:
    r"""Retrieve shift rule for a given quantum gate.
    """
    gate_name = gate_name.lower()
    return next(
        (
            (rule['shifts'], rule['coeffs'])
            for rule in SHIFT_RULES.values()
            if gate_name in rule['gates']
        ),
        ([DELTA, -DELTA], [0.5 / DELTA, -0.5 / DELTA]),
    )


_compute_measure: Callable[[StateOperator, List[int], int, str, Tuple[torch.Tensor, ...]], torch.Tensor] = OperatorMeasure.apply
_compute_expec_val_per_term: Callable[[StateOperator, Hamiltonian, torch.Tensor, Tuple[torch.Tensor, ...]], torch.Tensor] = OperatorExpecVal.apply
        



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
QuAIRKit
========

QuAIRKit is a Python SDK for algorithm development in quantum computing,
quantum information, and quantum machine learning. It focuses on flexible
design, real-time simulation, and rapid verification of quantum and classical
algorithms.

Links
-----
- Documentation: https://quairkit.com/QuAIRKit/latest/index.html
- PyPI: https://pypi.org/project/quairkit/
- Source: https://github.com/QuAIR/QuAIRKit
- License: Apache-2.0

Key features
------------
- Quantum algorithm simulation and optimization
- Quantum circuit simulation and visualization
- Quantum channel simulation
- Quantum algorithm/information tools
- Batch and probabilistic execution; qubit and qudit support
- Seamless integration with PyTorch tensors and autograd
- Optional third-party quantum cloud backends

Installation
------------
Minimum Python: 3.8 (recommended 3.10).

.. code-block:: bash

    pip install quairkit

Or using conda and an editable install:

.. code-block:: bash

    conda create -n quair python=3.10
    conda activate quair
    conda install jupyter notebook
    git clone https://github.com/QuAIR/QuAIRKit
    cd QuAIRKit
    pip install -e . --config-settings editable_mode=strict

Quick start
-----------
Import and basic setup:

.. code-block:: python

    import quairkit as qkit
    from quairkit import Circuit
    from quairkit.database import zero_state, pauli_group
    from quairkit.qinfo import state_fidelity

    # Global defaults
    qkit.set_dtype('complex128')  # default is 'complex64'
    qkit.set_device('cpu')        # or 'cuda' if available
    qkit.set_seed(73)

    # Build and run a simple circuit (batched over I/X/Y/Z)
    cir = Circuit(1)
    cir.oracle(pauli_group(1), 0)
    cir.ry(param=[0.0, 1.0, 2.0, 3.0])
    out = cir(zero_state(1))
    fid = state_fidelity(out, zero_state(1))

Qudit computation
-----------------
Work with heterogeneous system dimensions:

.. code-block:: python

    from quairkit.database import heisenberg_weyl, h

    cir = Circuit(2, system_dim=[2, 3])  # qubit + qutrit
    cir.oracle(heisenberg_weyl(6), [0, 1])
    cir.oracle(h(), [1, 0], control_idx=0)  # H on qubit, controlled by qutrit
    rho_qubit = cir().trace(1)  # trace out qutrit

Probabilistic / LOCC workflows
------------------------------
Model measurement, post-selection, and LOCC:

.. code-block:: python

    import torch
    from quairkit.database import eye, x, z, bell_state, random_state, nkron
    from quairkit.qinfo import state_fidelity

    M1 = torch.stack([eye(), x()])  # apply X for outcome 1
    M2 = torch.stack([eye(), z()])  # apply Z for outcome 1

    cir = Circuit(3)
    cir.cnot([0, 1])
    cir.h(0)
    cir.locc(M1, [1, 2])
    cir.locc(M2, [0, 2])

    psi = random_state(1, size=100)
    inp = nkron(psi, bell_state(2))
    out = cir(inp).trace([0, 1]).expec_state()
    avg_fid = state_fidelity(out, psi).mean().item()

Visualization
-------------
Render circuits via LaTeX/Quantikz:

.. code-block:: python

    cir: Circuit = ...
    cir.plot(print_code=True)

Cloud backends
--------------
Integrate third-party quantum cloud providers for execution:

.. code-block:: python

    class YourState(qkit.StateOperator):
        def _execute(self, qasm: str, shots: int):
            '''Execute the qasm on your backend and return a counts dict.'''

    qkit.set_backend(YourState)

Noise and mixed-state tools
---------------------------
Backends are selected implicitly based on operations:

.. code-block:: python

    cir = Circuit(3)
    cir.complex_entangled_layer(depth=3)
    _ = cir()                     # state_vector backend
    _ = cir().transpose([0, 1])   # switches to density_matrix
    cir.depolarizing(prob=0.1)
    _ = cir()                     # stays density_matrix

Module overview
---------------
- quairkit.circuit: Quantum circuit interface (torch.nn.Module)
- quairkit.database: Common matrices, sets, Hamiltonians, states, etc.
- quairkit.qinfo: Quantum information and algorithm utilities
- quairkit.loss: Loss operators for training
- quairkit.ansatz: Layer templates and ansätze

Tutorials
---------
For notebooks and examples, see:

- https://quairkit.com/QuAIRKit/latest/index.html
- https://github.com/QuAIR/QuAIRKit/tree/main/tutorials
"""

import os
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

from .core import set_backend, set_device, set_dtype, set_seed
from .core import get_backend, get_device, get_dtype, get_seed, get_float_dtype
from .core import Hamiltonian, State, Operator, to_state, StateOperator
from . import ansatz
from . import database
from . import operator
from . import loss
from . import qinfo
from .circuit import Circuit
from . import application

name = "quairkit"
__version__ = "0.4.1"


def print_info() -> None:
    r"""Print the information of QuAIRKit, its dependencies and current environment.

    """
    import torch
    import numpy
    import scipy
    import matplotlib
    print("\n---------VERSION---------")
    print("quairkit:", __version__)
    print("torch:", torch.__version__)
    if torch.cuda.is_available():
        print("torch cuda:", torch.version.cuda)
    print("numpy:", numpy.__version__)
    print("scipy:", scipy.__version__)
    print("matplotlib:", matplotlib.__version__)

    import platform
    print("---------SYSTEM---------")
    print("Python version:", platform.python_version())
    print("OS:", platform.system())
    print("OS version:", platform.version())

    
    import subprocess
    import re
    # stack overflow #4842448
    print("---------DEVICE---------")
    if platform.system() == "Windows":
        cpu_name = platform.processor()
        
    elif platform.system() == "Darwin":
        model = subprocess.check_output(["/usr/sbin/sysctl", "-n", "machdep.cpu.brand_string"]).strip()
        cpu_name = model.decode('utf-8')

    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).decode().strip()
        for line in all_info.split("\n"):
            if "model name" in line:
                cpu_name = re.sub(".*model name.*:", "", line, 1)
    print("CPU:", cpu_name)

    if torch.cuda.is_available():
        print("GPU: (0)", torch.cuda.get_device_name(0))
        for i in range(torch.cuda.device_count() - 1):
            print(f"     ({i})", torch.cuda.get_device_name(i))



# !/usr/bin/env python3
# Copyright (c) 2023 QuAIR team. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Common functions for the State class.
"""

import warnings
from functools import reduce
from typing import List, Optional, Type, Union

import numpy as np
import torch

from ..base import get_device, get_dtype
from ..operator import OperatorInfo, OperatorInfoType
from .backend import (DefaultSimulator, State, StateOperator, StateSimulator,
                      get_backend)


def __to_simulator_state(
    initializer: Type[StateSimulator],
    data: Union[torch.Tensor, np.ndarray, State],
    system_dim: Union[int, List[int]],
    eps: Optional[float],
    prob: Optional[List[torch.Tensor]],
) -> Union[DefaultSimulator, StateSimulator]:
    r"""Convert the input data to a state simulator."""
    if issubclass(initializer, StateOperator):
        raise NotImplementedError(
            f"Cannot convert numerical data to a state when the backend {initializer.backend} is not a simulator. "
            "Please set the backend to a simulator first, e.g., `set_backend('default')`."
        )

    dtype, device = get_dtype(), get_device()
    data = (
        data.to(dtype=dtype, device=device)
        if isinstance(data, torch.Tensor)
        else torch.tensor(data, dtype=dtype, device=device)
    )

    num_systems = initializer.check(data, system_dim, eps)
    system_dim = (
        [system_dim] * num_systems if isinstance(system_dim, int) else system_dim
    )
    return initializer(data, system_dim, list(range(num_systems)), prob)


def __to_operator_state(
    initializer: Type[StateOperator],
    data: List[OperatorInfoType],
    system_dim: Union[int, List[int]],
    eps: Optional[float],
    prob: Optional[List[torch.Tensor]],
) -> StateOperator:
    r"""Convert the input data to a state operator."""
    if eps != 1e-4:
        warnings.warn(
            "Error tolerance `eps` cannot used for the state operator that inputs a list of operators. "
        )

    if prob:
        warnings.warn(
            "Probability distribution `prob` cannot used for the state operator that inputs a list of operators. "
        )

    if issubclass(initializer, StateSimulator):
        raise NotImplementedError(
            f"Cannot convert a state to a operator when the backend {initializer.backend} is not a operator. "
        )

    if data is None:
        return initializer(data, system_dim)

    assert isinstance(data, List) and all(
        isinstance(op, OperatorInfo) for op in data
    ), f"Data should be a list of OperatorInfo, but got {type(data)}."

    return initializer(data, system_dim)


def to_state(
    data: Union[torch.Tensor, np.ndarray, State, Optional[List[OperatorInfoType]]],
    system_dim: Union[int, List[int]] = 2,
    eps: Optional[float] = 1e-4,
    backend: Optional[str] = None,
    prob: Optional[List[torch.Tensor]] = None,
) -> Union[DefaultSimulator, State]:
    r"""The function to generate a specified state instance.

    Args:
        data: a representation of the quantum state in allowable backend, or an instance of the State class.
        system_dim: (list of) dimension(s) of the systems, can be a list of integers or an integer.
         For example, ``system_dim = 2`` means the systems are qubits (default setup); ``system_dim = [2, 3]`` means
         the first system is a qubit and the second is a qutrit.
        eps: The tolerance for checking the validity of the input state. Can be adjusted to ``None`` to disable the check.
        backend: The name of the backend to use. If not specified, the default backend will be used.
        prob: The (list of) probability distribution of the state. The length of the list denotes the number of distributions.

    Returns:
        The generated quantum state.

    """
    if isinstance(data, State):
        if backend and data.backend != backend:
            raise NotImplementedError(
                f"Cannot convert a state to another backend. The current backend is {data.backend}, "
                f"but the target backend is {backend}. Please set the backend to {data.backend} first."
            )

        new_state = data.clone()
        system_dim = (
            [system_dim] * new_state.num_systems
            if isinstance(system_dim, int)
            else system_dim
        )
        new_state.system_dim = system_dim
        return new_state

    initializer = get_backend(backend)
    if isinstance(data, (torch.Tensor, np.ndarray)):
        return __to_simulator_state(initializer, data, system_dim, eps, prob)
    return __to_operator_state(initializer, data, system_dim, eps, prob)


def tensor_state(state_1st: StateSimulator, *args: StateSimulator) -> StateSimulator:
    r"""calculate tensor product (kronecker product) between at least two state. This function automatically returns State instance

    Args:
        state_1st: the first state
        args: other states

    Returns:
        tensor product state of input states

    Note:
        Need to be careful with the backend of states; Support broadcasting for batch states.
        Use ``quairkit.linalg.nkron`` if the input datatype is ``torch.Tensor`` or ``numpy.ndarray``.

    """
    __kron_state = lambda state_1st, state_2nd: state_1st.kron(state_2nd)
    return reduce(__kron_state, args, state_1st) if args else state_1st
